/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "agent.h"
#include "channel.h"
#include "network_constants.h"
#include "network_utils.h"
#include "serialization.h"
#include "utils.h"

static void yami_common_init(struct yami_agent * const self,
    struct yami_channel * channels_array, size_t channels_array_size)
{
    size_t i;
    
    self->channels_array_ = channels_array;
    self->channels_array_size_ = channels_array_size;
    
    for (i = 0u; i != self->channels_array_size_; ++i)
    {
        /* branch 1 */
        yami_channel_init(&self->channels_array_[i]);
    }
    
    self->listening_socket_ = INVALID_FILE_DESCRIPTOR;
    
    self->last_core_message_id_ = 0u;
    self->last_message_id_ = 0LL;
    
    self->raw_message_callback_ = NULL;
    self->message_callback_ = NULL;
    self->io_error_callback_ = NULL;
    self->connection_callback_ = NULL;
}

/* helper, finds the unused channel slot */
static enum yami_result yami_find_free_channel(
    const struct yami_agent * const self, size_t * const out_indx)
{
    enum yami_result res = yami_not_enough_space;
    
    size_t i = 0u;
    
    while (i != self->channels_array_size_)
    {
        /* branch 1 */
        if (yami_channel_is_open(&self->channels_array_[i]) == 0)
        {
            /* branch 1_c1 */
            *out_indx = i;
            res = yami_ok;
            
            break;
        }
        
        ++i;
    }
    
    return res;
}

/* helper, finds the given channel slot */
static enum yami_result yami_find_channel(
    const struct yami_agent * const self,
    const char target[], size_t * const out_indx)
{
    enum yami_result res = yami_no_such_name;
    
    size_t i = 0u;
    
    while (i != self->channels_array_size_)
    {
        /* branch 1 */
        if (yami_channel_name_cmp(&self->channels_array_[i], target) == 0)
        {
            /* branch 1_c1 */
            *out_indx = i;
            res = yami_ok;
            
            break;
        }
        
        ++i;
    }
    
    return res;
}

void yami_init(struct yami_agent * const self)
{
    yami_common_init(self, self->channels_, MAX_NUMBER_OF_CHANNELS);
    
    yami_options_init(&self->options_);
}

void yami_init_with_user_array(struct yami_agent * self,
    struct yami_channel channels_array[], size_t channels_array_size)
{
    yami_common_init(self, channels_array, channels_array_size);
    
    yami_options_init(&self->options_);
}

void yami_init_with_options(struct yami_agent * const self,
    const struct yami_options * const options)
{
    yami_common_init(self, self->channels_, MAX_NUMBER_OF_CHANNELS);

    self->options_ = *options;
}

void yami_init_with_options_ua(struct yami_agent * self,
    const struct yami_options * options,
    struct yami_channel channels_array[], size_t channels_array_size)
{
    yami_common_init(self, channels_array, channels_array_size);

    self->options_ = *options;
}

enum yami_result yami_clean(struct yami_agent * const self)
{
    enum yami_result res = yami_ok;
    
    if (self->listening_socket_ != INVALID_FILE_DESCRIPTOR)
    {
        /* branch c1 */
        res = yami_clean_listener(self->listening_socket_,
            self->io_error_callback_);
    }
    
    if (res == yami_ok)
    {
        /* branch c2 */
        size_t i;
        
        self->listening_socket_ = INVALID_FILE_DESCRIPTOR;;
        
        /* clean all used channels */
        for (i = 0u; i != self->channels_array_size_; ++i)
        {
            /* branch c2_1 */
            struct yami_channel * const ch = &self->channels_array_[i];
            
            const int32_t is_open = yami_channel_is_open(ch);
            
            if (is_open != 0)
            {
                /* branch c2_1_c1 */
                res = yami_channel_close(ch, self->io_error_callback_);
                if (res != yami_ok)
                {
                    /* branch c2_1_c1_c1 */
                    break;
                }
            }
        }
    }
    
    return res;
}

void yami_set_io_error_callback(struct yami_agent * const self,
    const yami_io_error_function io_error_callback)
{
    self->io_error_callback_ = io_error_callback;
}

void yami_set_raw_message_callback(struct yami_agent * const self,
    const yami_raw_message_function raw_message_callback)
{
    self->raw_message_callback_ = raw_message_callback;
}

void yami_set_message_callback(struct yami_agent * const self,
    const yami_message_function message_callback)
{
    self->message_callback_ = message_callback;
}

void yami_set_connection_callback(struct yami_agent * const self,
    const yami_connection_event_function connection_callback)
{
    self->connection_callback_ = connection_callback;
}

enum yami_result yami_set_listener(struct yami_agent * const self,
    const char target[])
{
    enum yami_result res;
    
    if (self->listening_socket_ != INVALID_FILE_DESCRIPTOR)
    {
        /* branch c1 */
        res = yami_bad_state;
    }
    else
    {
        /* branch c2 */
        int32_t socket;
        enum yami_protocol protocol;
        
        res = yami_create_listener(target, &self->options_,
            &socket, &protocol, self->io_error_callback_);
        if (res == yami_ok)
        {
            /* branch c2_c1 */
            self->listening_socket_ = socket;
            self->listening_protocol_ = protocol;
        }
    }
    
    return res;
}

static enum yami_result yami_ensure_connection(
    struct yami_agent * const self, const char target[],
    enum yami_connection_event event,
    size_t * const out_index)
{
    enum yami_result res;
    
    res = yami_find_channel(self, target, out_index);
    if (res == yami_no_such_name)
    {
        /* branch c1 */
        /* no such connection exists - create it */
        
        res = yami_find_free_channel(self, out_index);
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            res = yami_channel_open(&self->channels_array_[*out_index], target,
                &self->options_, self->io_error_callback_);
            
            if (res == yami_ok)
            {
                /* branch c1_c1_c1 */
                if (self->connection_callback_ != NULL)
                {
                    /* branch c1_c1_c1_c1 */
                    self->connection_callback_(event, target);
                }
            }
        }
    }
    
    return res;
}

enum yami_result yami_open_connection(struct yami_agent * const self,
    const char target[])
{
    size_t dummy_index;

    const enum yami_result res = yami_ensure_connection(
        self, target, yami_new_outgoing_connection, &dummy_index);
    
    return res;
}

enum yami_result yami_close_connection(struct yami_agent * const self,
    const char target[])
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_no_such_name)
    {
        /* branch c1 */
        /* no such connection exists - this is a success condition as well */

        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        res = yami_channel_close(&self->channels_array_[i], self->io_error_callback_);
        
        if (res == yami_ok)
        {
            /* branch c2_c1 */
            if (self->connection_callback_ != NULL)
            {
                /* branch c2_c1_c1 */
                self->connection_callback_(
                    yami_connection_closed, target);
            }
        }
    }
    
    return res;
}

enum yami_result yami_is_output_channel_busy(const struct yami_agent * const self,
    const char target[], int32_t * const out_busy)
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_ok)
    {
        /* branch c1 */
        *out_busy = self->channels_array_[i].output_buffer_has_data_;
    }
   
    return res;
}

enum yami_result yami_post_raw_message(struct yami_agent * const self,
    const char target[],
    const uint8_t header[], const size_t header_size,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_ok)
    {
        /* branch c1 */
        ++self->last_core_message_id_;
        
        res = yami_channel_post_raw_message(&self->channels_array_[i],
            self->last_core_message_id_,
            header, header_size, body, body_size);
    }
   
    return res;
}

enum yami_result yami_post_message(struct yami_agent * const self,
    const char target[], const char object_name[], const char message_name[],
    const uint8_t body[], const size_t body_size,
    int64_t * const out_message_id)
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_ok)
    {
        /* branch c1 */
        ++self->last_core_message_id_;
        
        ++self->last_message_id_;
        *out_message_id = self->last_message_id_;
        
        res = yami_channel_post_message(&self->channels_array_[i],
            object_name, message_name,
            self->last_message_id_, self->last_core_message_id_,
            body, body_size);
    }
   
    return res;
}

enum yami_result yami_post_reply(struct yami_agent * const self,
    const char target[], const int64_t message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_ok)
    {
        /* branch c1 */
        ++self->last_core_message_id_;
        
        res = yami_channel_post_reply(&self->channels_array_[i],
            message_id, self->last_core_message_id_,
            body, body_size);
    }
   
    return res;
}

enum yami_result yami_post_reject(struct yami_agent * const self,
    const char target[], const int64_t message_id, const char reason[])
{
    enum yami_result res;
    
    size_t i;
    
    res = yami_find_channel(self, target, &i);
    if (res == yami_ok)
    {
        /* branch c1 */
        ++self->last_core_message_id_;
        
        res = yami_channel_post_reject(&self->channels_array_[i],
            message_id, self->last_core_message_id_,
            reason);
    }
   
    return res;
}

/* helper, dispatches incoming message based on header information */
static enum yami_result yami_dispatch_incoming_message(
    const struct yami_agent * const self,
    struct yami_channel * const channel)
{
    enum yami_result res = yami_ok;
    
    const size_t body_size = channel->input_message_payload_size_ -
        channel->input_message_header_size_;

    if (self->raw_message_callback_ != NULL)
    {
        /* branch c1 */
        /* raw callback -> pass the header unparsed */
            
        self->raw_message_callback_(channel->target_,
            &channel->input_buffer_[FRAME_HEADER_SIZE],
            channel->input_message_header_size_,
            &channel->input_buffer_[FRAME_HEADER_SIZE + channel->input_message_header_size_],
            body_size);
    }
        
    if (self->message_callback_ != NULL)
    {
        /* branch c2 */
        /* high-level callback -> parse the message header */
            
        int64_t message_id;
        char header_type[MAX_HEADER_TYPE_LENGTH];
        char object_name[MAX_OBJECT_NAME_LENGTH];
        char message_name[MAX_MESSAGE_NAME_LENGTH];
        char reason[MAX_EXCEPTION_REASON_LENGTH];
            
        const char message_tag[] = "message";
        const char reply_tag[] = "reply";
        const char exception_tag[] = "exception";

        const char empty_message_name[1] = "";
        const char empty_object_name[1] = "";
        const char empty_exception_msg[1] = "";
        
        const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
        const size_t empty_body_size = sizeof(empty_body);

        size_t rpos = FRAME_HEADER_SIZE;
            
        res = yami_parse_message_header(
            channel->input_buffer_, FRAME_HEADER_SIZE + channel->input_message_header_size_, &rpos,
            header_type, MAX_HEADER_TYPE_LENGTH,
            object_name, MAX_OBJECT_NAME_LENGTH,
            message_name, MAX_MESSAGE_NAME_LENGTH,
            reason, MAX_EXCEPTION_REASON_LENGTH,
            &message_id);
            
        if (res == yami_ok)
        {
            /* branch c2_c1 */
            if (yami_strcmp(header_type, message_tag) == 0)
            {
                /* branch c2_c1_c1 */
                /* this is a regular incoming message */
                    
                self->message_callback_(channel->target_,
                    yami_incoming_message,
                    object_name, message_name, empty_exception_msg,
                    message_id,
                    &channel->input_buffer_[FRAME_HEADER_SIZE + channel->input_message_header_size_],
                    body_size);
            }
            else if (yami_strcmp(header_type, reply_tag) == 0)
            {
                /* branch c2_c1_c2 */
                /* this is a reply message */
                    
                self->message_callback_(channel->target_,
                    yami_message_reply,
                    empty_object_name, empty_message_name, empty_exception_msg,
                    message_id,
                    &channel->input_buffer_[FRAME_HEADER_SIZE + channel->input_message_header_size_],
                    body_size);
            }
            else if (yami_strcmp(header_type, exception_tag) == 0)
            {
                /* branch c2_c1_c3 */
                /* this is a rejection (exception) message */
                    
                self->message_callback_(channel->target_,
                    yami_message_rejected,
                    empty_object_name, empty_message_name, reason,
                    message_id,
                    empty_body, empty_body_size);
            }
            else
            {
                /* branch c2_c1_c4 */
                res = yami_unexpected_value;
            }
        }
    }

    if (res == yami_ok)
    {
        /* branch c3 */
        /* allow new messages to arrive on this channel */

        yami_channel_clean_input_buffer(channel);
    }
    
    return res;
}

static enum yami_result yami_accept_incoming_tcp_conn(
    struct yami_agent * const self)
{
    /* add new incoming connection to the pool */
    
    enum yami_result res;
    char target[MAX_LENGTH_OF_TARGET_NAME];
    int32_t sock;
    size_t i;

    /* the selector never reports the incoming connection                 */
    /* on the listening socket if there are no free channels in the pool, */
    /* the following call will always succeed                             */

    (void) yami_find_free_channel(self, &i);
    
    res = yami_accept_tcp(
        self->listening_socket_,
        &self->options_,
        target,
        &sock,
        self->io_error_callback_);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        yami_channel_init(&self->channels_array_[i]);
        
        self->channels_array_[i].socket_ = sock;
        self->channels_array_[i].protocol_ = yami_tcp;
        
        yami_strncpy(self->channels_array_[i].target_, target, MAX_LENGTH_OF_TARGET_NAME);
        
        if (self->connection_callback_ != NULL)
        {
            /* branch c1_c1 */
            self->connection_callback_(
                yami_new_incoming_connection, target);
        }
    }

    return res;
}

static enum yami_result yami_accept_incoming_udp_conn(
    struct yami_agent * const self)
{
    /* The UDP "listening" socket is ready to receive a packet       */
    /* but before that happens it is not clear whether it will       */
    /* be a new remote endpoint or a subsequent packet               */
    /* from the already known endpoint.                              */
    /* For this reason the packet has to be received and then        */
    /* the decision is taken whether a new logical channel has to be */
    /* created or an existing channel can be reused.                 */

    char target[MAX_LENGTH_OF_TARGET_NAME];
    uint8_t input_buffer[MAX_INCOMING_MESSAGE_PAYLOAD];
    size_t readn;

    const int32_t format_target = 1; /* true */

    enum yami_result res = yami_receive_udp(self->listening_socket_,
        target, format_target,
        input_buffer, MAX_INCOMING_MESSAGE_PAYLOAD, &readn,
        self->io_error_callback_);

    if (res == yami_ok)
    {
        /* branch c1 */
        /* make sure this channel already exists */

        size_t i;

        res = yami_ensure_connection(self, target,
            yami_new_incoming_connection, &i);
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            struct yami_channel * const channel = &self->channels_array_[i];

            /* inject the received packet into the channel */

            yami_memcpy(channel->input_buffer_, 0u, input_buffer, 0u, readn);

            channel->input_data_read_so_far_ = readn;

            /* allow the channel to interpret injected data */

            res = yami_channel_check_udp_packet(channel);
            if (res == yami_ok)
            {
                /* branch c1_c1_c1 */
                /* dispatch this incoming message and ignore message parsing errors */
                    
                (void) yami_dispatch_incoming_message(self, channel);
            }
            else
            {
                /* branch c1_c1_c2 */
                /* ignore packet format errors */

                res = yami_ok;
            }
        }
    }

    return res;
}

enum yami_result yami_do_some_work(struct yami_agent * const self,
    const size_t timeout)
{
    int32_t listening_sock_ready;
    size_t i;
    
    enum yami_result res = yami_wait_for_work(
        self->channels_array_, self->channels_array_size_,
        self->listening_socket_, &listening_sock_ready,
        timeout, self->io_error_callback_);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        if (listening_sock_ready != 0)
        {
            /* branch c1_c1 */
            if (self->listening_protocol_ == yami_tcp)
            {
                /* branch c1_c1_c1 */
                res = yami_accept_incoming_tcp_conn(self);
            }
            else /* yami_udp */
            {
                /* branch c1_c1_c2 */
                res = yami_accept_incoming_udp_conn(self);
            }
        }
    }

    if (res == yami_ok)
    {
        /* branch c2 */
        for (i = 0u; i != self->channels_array_size_; ++i)
        {
            /* branch c2_1 */
            struct yami_channel * const channel = &self->channels_array_[i];
            
            if (res == yami_ok)
            {
                /* branch c2_1_c1 */
                if (channel->socket_ready_for_input_ != 0)
                {
                    /* branch c2_1_c1_c1 */
                    res = yami_channel_read_some_data(channel,
                        self->io_error_callback_);
                    
                    if (res == yami_ok)
                    {
                        /* branch c2_1_c1_c1_c1 */
                        if (channel->input_message_complete_ != 0)
                        {
                            /* branch c2_1_c1_c1_c1_c1 */
                            /* dispatch this incoming message */
                            
                            res = yami_dispatch_incoming_message(self, channel);
                        }
                    }
                }
            }
            
            if (res == yami_ok)
            {
                /* branch c2_1_c2 */
                if (channel->socket_ready_for_output_ != 0)
                {
                    /* branch c2_1_c2_c1 */
                    res = yami_channel_write_some_data(channel,
                        self->io_error_callback_);
                }
            }
            
            /* decision channel_fault_when_doing_work: (res == yami_io_error) || (res == yami_channel_closed) || (res == yami_not_enough_space) */
            /*                                        A: res == yami_io_error         */
            /*                                        B: res == yami_channel_closed   */
            /*                                        C: res == yami_not_enough_space */
            /* end                                                                    */
            if ((res == yami_io_error) || (res == yami_channel_closed) || (res == yami_not_enough_space))
            {
                /* branch c2_1_c3 */
                /* abandon any pending outgoing message and close the faulty channel */
                
                if (channel->output_buffer_has_data_ != 0)
                {
                    /* branch c2_1_c3_c1 */
                    if (channel->output_message_id_ != 0LL)
                    {
                        /* branch c2_1_c3_c1_c1 */
                        if (self->message_callback_ != NULL)
                        {
                            /* branch c2_1_c3_c1_c1_c1 */
                            /* call message_callback with yami_message_abandoned */
                            
                            const char empty_message_name[1] = "";
                            const char empty_object_name[1] = "";
                            const char empty_exception_msg[1] = "";
                            
                            const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
                            const size_t empty_body_size = sizeof(empty_body);
                            
                            self->message_callback_(channel->target_,
                                yami_message_abandoned,
                                empty_object_name, empty_message_name, empty_exception_msg,
                                channel->output_message_id_,
                                empty_body, empty_body_size);
                        }
                    }
                }
                
                if (self->connection_callback_ != NULL)
                {
                    /* branch c2_1_c3_c2 */
                    self->connection_callback_(
                        yami_connection_closed, channel->target_);
                }

                (void)yami_channel_close(channel, self->io_error_callback_);
                
                if (res == yami_io_error)
                {
                    /* branch c2_1_c3_c3 */
                    break;
                }
                else /* (res == yami_channel_closed) || (res == yami_not_enough_space) */
                {
                    /* branch c2_1_c3_c4 */
                    /* this is not considered as error at this level */
                    res = yami_ok;
                }
            }
        }
    }
    
    return res;
}
