/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef YAMI4INDUSTRY_AGENT_H
#define YAMI4INDUSTRY_AGENT_H

#include "callbacks.h"
#include "channel.h"
#include "common.h"
#include "limits.h"
#include "options.h"

#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/** \brief Message broker.
 * 
 * The message broker that encapsulates physical channel management,
 * incoming and outgoing message buffers, optional listener and
 * related resource management activities.
 * 
 * The agent objects can be created and destroyed without constraints
 * on the stack, on the free store or as static objects.
 */
struct yami_agent
{
    int32_t listening_socket_;
    enum yami_protocol listening_protocol_;
    
    uint32_t last_core_message_id_;
    int64_t last_message_id_;
    
    yami_raw_message_function raw_message_callback_;
    yami_message_function message_callback_;
    yami_io_error_function io_error_callback_;
    yami_connection_event_function connection_callback_;
    
    struct yami_options options_;
    
    struct yami_channel * channels_array_;
    size_t channels_array_size_;
    
    struct yami_channel channels_[MAX_NUMBER_OF_CHANNELS];
};

/** \brief Initializer.
 * 
 * Initializes the agent object to the basic operational state,
 * using the internal static array of channels.
 * No network resources are allocated.
 * All socket-related options have their default values.
 *
 * This function always succeeds as it does not take into account any
 * previous object state.
 *
 * @param self this object
 */
void yami_init(struct yami_agent * self);

/** \brief Initializer.
 * 
 * Initializes the agent object to the basic operational state,
 * using the array of channels allocated by user;
 * channels do not have to be initialized.
 * No network resources are allocated.
 * All socket-related options have their default values.
 *
 * This function always succeeds as it does not take into account any
 * previous object state.
 *
 * @param self this object
 * @param channels_array array of channel structures, preallocated by caller
 * @param channels_array_size number of channel structures in the array
 */
void yami_init_with_user_array(struct yami_agent * self,
    struct yami_channel channels_array[], size_t channels_array_size);

/** \brief Initializer.
 * 
 * Initializes the agent object to the basic operational state
 * with the given set of options.
 * No network resources are allocated.
 *
 * This function always succeeds as it does not take into account any
 * previous object state.
 *
 * @param self this object
 * @param options user-provided option values
 */
void yami_init_with_options(struct yami_agent * self,
    const struct yami_options * options);

/** \brief Initializer.
 * 
 * Initializes the agent object to the basic operational state
 * with the given set of options and the array of channels allocated by user;
 * channels do not have to be initialized.
 * No network resources are allocated.
 *
 * This function always succeeds as it does not take into account any
 * previous object state.
 *
 * @param self this object
 * @param options options user-provided option values
 * @param channels_array array of channel structures, preallocated by caller
 * @param channels_array_size number of channel structures in the array
 */
void yami_init_with_options_ua(struct yami_agent * self,
    const struct yami_options * options,
    struct yami_channel channels_array[], size_t channels_array_size);

/** \brief Destructor.
 * 
 * Cleans up the agent's system resources.
 *
 * @param self this object
 *
 * @return <code>yami_ok</code> or <code>yami_io_error</code> in case
 *         of any system-related errors.
 */
enum yami_result yami_clean(struct yami_agent * self);

/** \brief Sets the error logger callback function.
 *
 * Sets the error logger callback function that will be called in case
 * of I/O or system error.
 *
 * This function always succeeds.
 *
 * @param self this object
 * @param io_error_callback user-provided I/O error logger callback
 */
void yami_set_io_error_callback(struct yami_agent * self,
    yami_io_error_function io_error_callback);

/** \brief Sets the raw message callback.
 *
 * Sets the raw message callback function, which will be called
 * when there is a new incoming message.
 *
 * Note: It is possible to execute any function from the callback,
 * but depending on the actual action it may lead to further
 * (that is, recursive) callback invocations.
 *
 * This function always succeeds.
 *
 * @param self this object
 * @param raw_message_callback user-provided raw message callback
 */
void yami_set_raw_message_callback(struct yami_agent * self,
    yami_raw_message_function raw_message_callback);

/** \brief Sets the high-level message callback.
 *
 * Sets the high-level message callback function, which will be called
 * when there is a new incoming message, or when there is a reply,
 * rejection or error report for previously sent outgoing message.
 *
 * Note: It is possible to execute any function from the callback,
 * but depending on the actual action it may lead to further
 * (that is, recursive) callback invocations.
 *
 * This function always succeeds.
 *
 * @param self this object
 * @param message_callback user-provided high-level message callback
 */
void yami_set_message_callback(struct yami_agent * self,
    yami_message_function message_callback);

/** \brief Sets the connection event callback.
 *
 * Sets the connection event callback function, which will be called
 * whenever new connection is created or existing connection is closed.
 *
 * Note: It is possible to execute any function from the callback,
 * but depending on the actual action it may lead to further
 * (that is, recursive) callback invocations.
 *
 * This function always succeeds.
 *
 * @param self this object
 * @param connection_callback user-provided connection callback
 */
void yami_set_connection_callback(struct yami_agent * self,
    yami_connection_event_function connection_callback);

/** \brief Sets the optional listener.
 *
 * Sets the optional listener in the agent.
 * The listener will operate based on the given target.
 *
 * @param self this object
 * @param target target defining the protocol and local network endpoint
 *
 * @return <code>yami_ok</code> or
 *         <code>yami_io_error</code> in case of system errors or
 *         <code>yami_bad_state</code> if this agent already has a listener
 */
enum yami_result yami_set_listener(struct yami_agent * self,
    const char target[]);

/** \brief Creates (opens) outgoing communication channel.
 *
 * Creates (opens) outgoing communication channel if it does not yet exist.
 *
 * @param self this object
 * @param target endpoint name for the new connection
 *
 * @return <code>yami_ok</code> if the connection was successful
 *                              or the channel already exists or
 *         <code>yami_not_enough_space</code> if there is no place for the new channels or
 *         <code>yami_io_error</code> in case of system error
 */
enum yami_result yami_open_connection(struct yami_agent * self,
    const char target[]);

/** \brief Closes communication channel.
 *
 * Closes communication channel.
 * This function can be used to close outgoing and incoming communication channels.
 *
 * @param self this object
 * @param target endpoint name for the channel that is to be closed
 *
 * @return <code>yami_ok</code> if the oporation was successful
 *                              or the channel does not exist or
 *         <code>yami_io_error</code> in case of system error
 */
enum yami_result yami_close_connection(struct yami_agent * self,
    const char target[]);

/** \brief Checks if there is space in the output channel.
 *
 * Checks if there is space in the output channel for the new message.
 *
 * @param self this object
 * @param target endpoint name for the channel that is to be checked
 * @param out_busy variable to be set to 1 if the channel is busy
 *
 * @return <code>yami_ok</code> if the oporation was successful or
 *         <code>yami_no_such_name</code> if there is no such channel
 */
enum yami_result yami_is_output_channel_busy(const struct yami_agent * self,
    const char target[], int32_t * out_busy);

/** \brief Posts the raw message.
 *
 * Posts the raw message that was already serialized by user.
 * The user is supposed to provide his own message header.
 * Note: the message is placed in the outgoing buffer,
 * but in order to physically send it to the remote recevier
 * the I/O work needs to be done on this agent.
 *
 * @param self this object
 * @param target endpoint name for the channel where the message should be sent
 * @param header already serialized header buffer
 * @param header_size size of the header buffer
 * @param body already serialized body buffer
 * @param body_size size of the body buffer
 *
 * @return <code>yami_ok</code> if the message was posted,
 *         <code>yami_no_such_name</code> if the given target does not
 *                                        reference any open channel or
 *         <code>yami_not_enough_space</code> if the message is bigger
 *                                            than the buffer space
 */
enum yami_result yami_post_raw_message(struct yami_agent * self,
    const char target[],
    const uint8_t header[], size_t header_size,
    const uint8_t body[], size_t body_size);

/** \brief Posts the high-level message.
 *
 * Posts the high-level message.
 * Note: the message is placed in the outgoing buffer,
 * but in order to physically send it to the remote recevier
 * the I/O work needs to be done on this agent.
 *
 * @param self this object
 * @param target endpoint name for the channel where the message should be sent
 * @param object_name name of logical message destination
 * @param message_name name of message
 * @param body already serialized body buffer
 * @param body_size size of the body buffer
 * @param out_message_id high-level message identifier, allocated by this agent
 *
 * @return <code>yami_ok</code> if the message was posted,
 *         <code>yami_no_such_name</code> if the given target does not
 *                                        reference any open channel or
 *         <code>yami_not_enough_space</code> if the message is bigger
 *                                            than the buffer space
 */
enum yami_result yami_post_message(struct yami_agent * self,
    const char target[], const char object_name[], const char message_name[],
    const uint8_t body[], size_t body_size,
    int64_t * out_message_id);

/** \brief Posts the high-level message reply.
 *
 * Posts the high-level message reply.
 * Note: the message reply is placed in the outgoing buffer,
 * but in order to physically send it to the remote recevier
 * the I/O work needs to be done on this agent.
 *
 * @param self this object
 * @param target endpoint name for the channel where the reply should be sent,
 *               in typical scenario this should be the originating endpoint
 * @param message_id high-level message identifier, allocated by remote sender
 * @param body already serialized body buffer
 * @param body_size size of the body buffer
 *
 * @return <code>yami_ok</code> if the message was posted,
 *         <code>yami_no_such_name</code> if the given target does not
 *                                        reference any open channel or
 *         <code>yami_not_enough_space</code> if the message is bigger
 *                                            than the buffer space
 */
enum yami_result yami_post_reply(struct yami_agent * self,
    const char target[], int64_t message_id,
    const uint8_t body[], size_t body_size);

/** \brief Posts the high-level message rejection.
 *
 * Posts the high-level message rejection.
 * Note: the message rejection is placed in the outgoing buffer,
 * but in order to physically send it to the remote recevier
 * the I/O work needs to be done on this agent.
 *
 * @param self this object
 * @param target endpoint name for the channel where the rejection should be sent,
 *               in typical scenario this should be the originating endpoint
 * @param message_id high-level message identifier, allocated by remote sender
 * @param reason exception message, intended as description of
 *               why the original message is rejected
 *
 * @return <code>yami_ok</code> if the message was posted,
 *         <code>yami_no_such_name</code> if the given target does not
 *                                        reference any open channel or
 *         <code>yami_not_enough_space</code> if the message is bigger
 *                                            than the buffer space
 */
enum yami_result yami_post_reject(struct yami_agent * self,
    const char target[], int64_t message_id, const char reason[]);

/** \brief Performs some unit of network-related activity.
 *
 * @param self this object
 * @param timeout maximum blocking time, 0u means indefinite
 *
 * @return <code>yami_ok</code> if the work is performed,
 *         <code>yami_timed_out</code> if no work was pending during the given time,
 *         <code>yami_io_error</code> in case of system-related error,
 *         <code>yami_unexpected_value</code> if the new connection is created
 *                                            but the target cannot be formatted or
 *         <code>yami_bad_state</code> if there is no event to wait for
 */
enum yami_result yami_do_some_work(struct yami_agent * self,
    size_t timeout);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* YAMI4INDUSTRY_AGENT_H */
