/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "messaging.h"
#include "broker_limits.h"
#include "configuration.h"
#include "log.h"
#include "routing.h"

#include "../agent.h"
#include "../serialization.h"
#include "../utils.h"

static struct yami_agent agent_;
static enum subscription_overflow_policy overflow_policy_;

/* helper for sending back message confirmations (replies) */
static void confirm(const enum log_module module, const int64_t message_id, const char target[])
{
    const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
    const size_t empty_body_size = sizeof(empty_body);

    const enum yami_result res =
        yami_post_reply(&agent_, target, message_id, empty_body, empty_body_size);
    
    if (res == yami_ok)
    {
        log_put_ss(module, "    confirmed to ", target);
    }
    else
    {
        log_put(module, "cannot post confirmation - ignoring it");
    }
}

/* helper for sending back rejection in case of subscription overflow */
/* or when the subscription message itself is ill-formed              */
static void reject(const int64_t message_id,
    const char target[], const char reason[])
{
    const enum yami_result res =
        yami_post_reject(&agent_, target, message_id, reason);
    
    if (res == yami_ok)
    {
        log_put_ss(module_messages, "rejected message from ", target);
    }
    else
    {
        log_put(module_messages, "cannot post rejection - ignoring it");
    }
}

/* helper for sending the given message to the matching subscriber */
static enum yami_result forward_to_single_subscriber(
    const char subscriber_target[], const char subscriber_object_name[],
    const char message_name[],
    const uint8_t body[], const size_t body_size,
    int32_t * const out_channel_is_busy)
{
    enum yami_result res;
    int64_t forwarded_message_id;
    
    log_put_ss(module_messages, "    routing to ", subscriber_target);

    *out_channel_is_busy = 0; /* false until found otherwise */
    
    /* make sure that the forwarding channel exists */

    res = yami_open_connection(&agent_, subscriber_target);
    if (res == yami_ok)
    {
        res = yami_is_output_channel_busy(&agent_, subscriber_target, out_channel_is_busy);
        if ((res == yami_ok) && (*out_channel_is_busy == 0))
        {
            res = yami_post_message(&agent_, subscriber_target, subscriber_object_name,
                message_name, body, body_size, &forwarded_message_id);
        }
    }

    return res;
}

/* helper, processes the publish command by replicating the same */
/* message content to all matching subscribers                   */
static void process_publish(const int64_t message_id,
    const char target[], const char tags[],
    const uint8_t body[], const size_t body_size,
    const int32_t requires_confirmation)
{
    enum yami_result res;

    char message_name[MAX_MESSAGE_NAME_LENGTH];
    const char update_tag[] = "update ";
    const size_t update_tag_len = sizeof(update_tag) - 1u;

    char subscriber_target[MAX_TAG_PATTERN_LENGTH];
    char subscriber_object_name[MAX_TAG_PATTERN_LENGTH];
    size_t index;
    int32_t is_forwarding;
    int32_t channel_is_busy;
    int32_t overflow = 0; /* false */
    int32_t should_unsubscribe = 0;
    
    log_put_ssss(module_messages, "processing ", tags, " from ", target);

    /* for regular subscriptions                           */
    /* the message name should have the form "update TAGS" */

    yami_strncpy(message_name, update_tag, MAX_MESSAGE_NAME_LENGTH);
    yami_strncpy(&message_name[update_tag_len], tags,
        MAX_MESSAGE_NAME_LENGTH - update_tag_len);

    for (index = 0u; index != MAX_NUM_OF_SUBSCRIPTIONS; ++index)
    {
        const int32_t matches = routing_check_matching_entry(
            tags, index, subscriber_target, subscriber_object_name);

        if (matches != 0)
        {
            /* a matching subscription was found, */
            /* forward message to that subscriber */

            should_unsubscribe = 0; /* false */

            if (yami_strcmp(subscriber_object_name, "*") == 0)
            {
                /* this is a forwarding subscription          */
                /* use original object name (list of targets) */
                /* and "publish" as the message name          */

                is_forwarding = 1; /* true */

                res = forward_to_single_subscriber(
                    subscriber_target, tags, "publish",
                    body, body_size, &channel_is_busy);
            }
            else
            {
                /* this is a regular subscription */

                is_forwarding = 0; /* false */

                res = forward_to_single_subscriber(
                    subscriber_target, subscriber_object_name, message_name,
                    body, body_size, &channel_is_busy);
            }

            if (channel_is_busy != 0)
            {
                overflow = 1; /* true */

                if (overflow_policy_ == unsubscribe)
                {
                    should_unsubscribe = 1; /* true */

                    log_put_ss(module_subscriptions,
                        "queue full, abandoned subscription to ", subscriber_target);
                }
                else
                {
                    log_put_ss(module_subscriptions,
                        "queue full, dropped update for ", subscriber_target);
                }
            }

            if (res != yami_ok)
            {
                /* in case of I/O errors, silently close the channel           */
                /* and remove the subscriber from subscription list,           */
                /* unless it is a forwarding subscription, which are permanent */

                (void)yami_close_connection(&agent_, subscriber_target);

                if (is_forwarding == 0)
                {
                    log_put_ss(module_subscriptions,
                        "abandoned subscription to ", subscriber_target);

                    should_unsubscribe = 1; /* true */
                }
            }

            if (should_unsubscribe != 0)
            {
                routing_unsubscribe_at_index(index);
            }
        }
    }
    
    if (requires_confirmation == 1)
    {
        if ((overflow == 1) && (overflow_policy_ == reject_message))
        {
            reject(message_id, target, "message queue overflow");
        }
        else
        {
            confirm(module_messages, message_id, target);
        }
    }
}

static void process_subscribe(const int64_t message_id,
    const char target[], const char tags[],
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;

    int32_t num_of_entries;
    int32_t len;
    enum yami_parameter_type type;
    const char expected_field_name[] = "destination_object";
    const size_t expected_field_name_len = sizeof(expected_field_name) - 1u;
    char field_name[20]; /* enough for the expected name */

    char destination_obj[MAX_OBJECT_NAME_LENGTH];
    int32_t destination_obj_len;

    /* for simplicity expect only one string parameter "destination_object" */
    /* that is, do not allow unknown extensions of the message              */

    /* buffer position is updated with each reading operation */
    size_t buffer_position = 0u;
    
    res = yami_get_integer(body, body_size, &buffer_position, &num_of_entries);
    if ((res == yami_ok) && (num_of_entries != 1))
    {
        res = yami_unexpected_value;
    }
    
    if (res == yami_ok)
    {
        /* read length of field name */
        
        res = yami_get_integer(body, body_size, &buffer_position, &len);
        if ((res == yami_ok) && (len != (int32_t)expected_field_name_len))
        {
            res = yami_unexpected_value;
        }
    }
    
    if (res == yami_ok)
    {
        /* read the field name */
    
        res = yami_get_raw_cstring(body, body_size, &buffer_position,
            field_name, (size_t)len);
        if (res == yami_ok)
        {
            if (yami_strcmp(field_name, expected_field_name) != 0)
            {
                res = yami_unexpected_value;
            }
        }
    }
    
    if (res == yami_ok)
    {
        /* read type (expect string) */

        res = yami_get_type(body, body_size, &buffer_position, &type);
        if ((res == yami_ok) && (type != yami_string))
        {
            res = yami_unexpected_value;
        }
    }
    
    if (res == yami_ok)
    {
        /* read the field value length */

        res = yami_get_integer(body, body_size, &buffer_position,
            &destination_obj_len);
        if (res == yami_ok)
        {
            if ((destination_obj_len <= 0) ||
                (destination_obj_len > (int32_t)MAX_OBJECT_NAME_LENGTH))
            {
                res = yami_unexpected_value;
            }
        }
    }
    
    if (res == yami_ok)
    {
        /* read the field value */

        res = yami_get_raw_cstring(body, body_size, &buffer_position,
            destination_obj, (size_t)destination_obj_len);
    }

    if (res == yami_ok)
    {
        const int32_t subs_res = routing_subscribe(tags, destination_obj, target);
        if (subs_res == 1)
        {
            /* subscription was successful */

            confirm(module_subscriptions, message_id, target);
        }
        else
        {
            /* subscription failed (subscription table is already full) */

            reject(message_id, target, "cannot subscribe due to overflow");
        }
    }
    else
    {
        reject(message_id, target, "no destination object defined for subscription");
    }
}

/* main callback function that receives all incoming messages                     */
/* (subscription requests as well as messages that will be pushed to subscribers) */
static void incoming_message_handler(
    const char target[],
    const enum yami_message_type message_type,
    const char object_name[], const char message_name[], const char exception_msg[],
    const int64_t message_id,
    const uint8_t body[], const size_t body_size)
{
    int32_t requires_confirmation = 0; /* false */
    
    (void)exception_msg; /* intentionally unreferenced */
    
    if (message_type == yami_incoming_message)
    {
        if (yami_strcmp(message_name, "publish") == 0)
        {
            /* treat object name as a list of tags        */
            /* this message does not require confirmation */

            process_publish(message_id, target, object_name, body, body_size, requires_confirmation);
        }
        else if (yami_strcmp(message_name, "publish_confirm") == 0)
        {
            /* treat object name as a list of tags */
            /* this message requires confirmation  */

            requires_confirmation = 1;

            process_publish(message_id, target, object_name, body, body_size, requires_confirmation);
        }
        else if (yami_strcmp(message_name, "subscribe") == 0)
        {
            /* treat object name as a list of tags */

            process_subscribe(message_id, target, object_name, body, body_size);
        }
        else
        {
            /* unknown message - ignore it */
        }
    }
}

static enum yami_result init_agent(const struct broker_properties * const properties)
{
    enum yami_result res;
    
    yami_init(&agent_);
    
    yami_set_message_callback(&agent_, &incoming_message_handler);

    res = yami_set_listener(&agent_, properties->listener);
    if (res == yami_ok)
    {
        log_put_ss(module_main, "listening on ", properties->listener);
    }
    else
    {
        log_put_ss(module_main, "ERROR: cannot add listener: ", properties->listener);
    }
    
    return res;
}

void messaging_run(const struct broker_properties * const properties)
{
    enum yami_result res;
    
    res = init_agent(properties);
    if (res == yami_ok)
    {
        overflow_policy_ = properties->overflow_policy;

        while (1)
        {
            res = yami_do_some_work(&agent_, 1000u /* 1000ms timeout */);
        
            if ((res != yami_ok) && (res != yami_timed_out))
            {
                log_put(module_messages, "ERROR: run-time error in the event loop");

                break;
            }
        }
    }
}
