/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "routing.h"
#include "broker_limits.h"
#include "configuration.h"
#include "log.h"
#include "patterns.h"

#include "../utils.h"

/* The subscription type handles both regular subscriptions
 * as well as forward channels.
 * Notes about forward channels:
 * 1. forward channels are never removed
 *    (not even after failure or overflow)
 * 2. the original list of tags in the message
 *    is preserved as object name when the message is forwarded
 */
struct subscription
{
    int32_t active; /* boolean */
    
    char tag_patterns[MAX_TAG_PATTERN_LENGTH];
    char target_object[MAX_TAG_PATTERN_LENGTH];
    char target_location[MAX_TAG_PATTERN_LENGTH];
};

static struct subscription subscription_table[MAX_NUM_OF_SUBSCRIPTIONS];

static void subscription_init(struct subscription * const self)
{
    self->active = 0;
}

static void subscription_activate(struct subscription * const self,
    const char tags[],
    const char destination_obj[],
    const char source[])
{
    self->active = 1; /* true */
    
    yami_strncpy(self->tag_patterns, tags, MAX_TAG_PATTERN_LENGTH);
    
    yami_strncpy(self->target_object, destination_obj, MAX_TAG_PATTERN_LENGTH);
    
    yami_strncpy(self->target_location, source, MAX_TAG_PATTERN_LENGTH);
}

static void subscription_deactivate(struct subscription * const self)
{
    self->active = 0; /* false */
}

void routing_init()
{
    size_t i;
    
    for (i = 0u; i != MAX_NUM_OF_SUBSCRIPTIONS; ++i)
    {
        subscription_init(&subscription_table[i]);
    }
}

static int32_t is_duplicate(
    const struct subscription * const s,
    const char tags[],
    const char destination_obj[],
    const char source[])
{
    const int32_t tags_match = yami_strcmp(s->tag_patterns, tags);
    const int32_t obj_match = yami_strcmp(s->target_object, destination_obj);
    const int32_t target_match = yami_strcmp(s->target_location, source);
    
    return (tags_match == 0) && (obj_match == 0) && (target_match == 0);
}

int32_t routing_subscribe(
    const char tags[],
    const char destination_obj[],
    const char source[])
{
    int32_t res;

    size_t insert_point;
    int32_t found_insert_point = 0; /* false */
    int32_t found_duplicate = 0; /* false */
    
    struct subscription * sub;
    size_t i;
    
    for (i = 0u; i != MAX_NUM_OF_SUBSCRIPTIONS; ++i)
    {
        sub = &subscription_table[i];
        
        if (sub->active == 0)
        {
            if (found_insert_point == 0)
            {
                insert_point = i;
                found_insert_point = 1; /* true */
            }
        }
        else
        {
            /* do not duplicate identical subscriptions */
            
            if (is_duplicate(sub, tags, destination_obj, source))
            {
                log_put_sss(module_subscriptions,
                    sub->target_location, " refreshed subscription to ", sub->tag_patterns);
                
                found_duplicate = 1; /* true */
                break;
            }
        }
    }
    
    if (found_duplicate == 0)
    {
        if (found_insert_point == 1)
        {
            sub = &subscription_table[insert_point];

            subscription_activate(sub, tags, destination_obj, source);

            if (yami_strcmp(destination_obj, "*") == 0)
            {
                log_put_ssss(module_subscriptions,
                    "forwarding ", sub->tag_patterns, " to ", sub->target_location);
            }
            else
            {
                log_put_sss(module_subscriptions,
                    sub->target_location, " subscribed to ", sub->tag_patterns);
            }

            res = 1; /* true */
        }
        else
        {
            log_put_ss(module_subscriptions,
                source, " cannot subscribe due to overflow");

            res = 0; /* false */
        }
    }
    else
    {
        res = 1; /* true, existing entry counts as success */
    }

    return res;
}

int32_t routing_check_matching_entry(
    const char tags[],
    const size_t index,
    char out_target[], char out_object_name[])
{
    int32_t result = 0; /* false */

    if (index < MAX_NUM_OF_SUBSCRIPTIONS)
    {
        const struct subscription * const sub = &subscription_table[index];

        if (sub->active == 1)
        {
            if (patterns_multi_hierarchic_match(tags, sub->tag_patterns) == 1)
            {
                yami_strncpy(out_target, sub->target_location, MAX_TAG_PATTERN_LENGTH);
                yami_strncpy(out_object_name, sub->target_object, MAX_TAG_PATTERN_LENGTH);

                result = 1; /* true */
            }
        }
    }

    return result;
}

void routing_unsubscribe_at_index(const size_t index)
{
    subscription_deactivate(&subscription_table[index]);
}
