/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "channel.h"
#include "network_constants.h"
#include "network_utils.h"
#include "serialization.h"
#include "utils.h"

void yami_channel_init(struct yami_channel * const self)
{
    self->socket_ = INVALID_FILE_DESCRIPTOR;
    
    self->socket_ready_for_output_ = 0; /* false */
    self->socket_ready_for_input_ = 0; /* false */
    self->output_buffer_has_data_ = 0; /* false */
    self->input_buffer_has_space_ = 1; /* true */
    self->input_data_read_so_far_ = 0u;
    self->input_message_complete_ = 0; /* false */

    self->target_[0] = (char)'\0';
}

int32_t yami_channel_is_open(const struct yami_channel * const self)
{
    int32_t res;
    
    if (self->socket_ != INVALID_FILE_DESCRIPTOR)
    {
        /* branch c1 */
        res = 1;
    }
    else
    {
        /* branch c2 */
        res = 0;
    }
    
    return res;
}

int32_t yami_channel_name_cmp(
    const struct yami_channel * const self, const char target[])
{
    int32_t res;

    res = yami_strcmp(self->target_, target);
    
    return res;
}

enum yami_result yami_channel_open(struct yami_channel * const self,
    const char target[],
    const struct yami_options * const options,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res = yami_bad_state;
    
    if (self->socket_ == INVALID_FILE_DESCRIPTOR)
    {
        /* branch c1 */
        int32_t sock;
        enum yami_protocol protocol;
        uint32_t host;
        uint16_t port;

        res = yami_create_outgoing_connection(
            target, options, &sock, &protocol, &host, &port,
            io_error_callback);
        
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            self->socket_ = sock;
            self->protocol_ = protocol;
            self->host_ = host;
            self->port_ = port;
            
            yami_strncpy(self->target_, target, MAX_LENGTH_OF_TARGET_NAME);
        }
    }

    return res;
}

enum yami_result yami_channel_close(struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res = yami_bad_state;
    
    if (self->socket_ != INVALID_FILE_DESCRIPTOR)
    {
        /* branch c1 */
        res = yami_clean_connection(self->socket_, io_error_callback);
        
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            yami_channel_init(self);
        }
    }

    return res;
}

enum yami_result yami_channel_post_raw_message(struct yami_channel * const self,
    const uint32_t core_message_id,
    const uint8_t header[], const size_t header_size,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    if (self->output_buffer_has_data_ == 0)
    {
        /* branch c1 */
        /* copy buffers */
        
        const size_t frame_payload_size = header_size + body_size;
            
        if ((FRAME_HEADER_SIZE + frame_payload_size) <= MAX_OUTGOING_MESSAGE_PAYLOAD)
        {
            /* branch c1_c1 */
            yami_memcpy(self->output_buffer_, FRAME_HEADER_SIZE, header, 0u, header_size);
            yami_memcpy(self->output_buffer_, FRAME_HEADER_SIZE + header_size, body, 0u, body_size);
            
            yami_fill_frame_header(self->output_buffer_,
                core_message_id, header_size, frame_payload_size);
            
            self->output_buffer_has_data_ = 1; /* true, output buffer busy */
            self->output_data_size_ = FRAME_HEADER_SIZE + frame_payload_size;
            self->output_data_written_so_far_ = 0u;
            self->output_message_id_ = 0LL;

            res = yami_ok;
        }
        else
        {
            /* branch c1_c2 */
            res = yami_not_enough_space;
        }
    }
    else
    {
        /* branch c2 */
        /* output buffer is busy */
        
        res = yami_not_enough_space;
    }
    
    return res;
}

/* helper function that is a common part for yami_channel_post_message, */
/* yami_channel_post_reply and yami_channel_post_reject                 */
/* - fills the buffer with frame header and message body                */
/* and complete channel accounting for the new outgoing frame           */
static enum yami_result yami_complete_msg_preparation(struct yami_channel * const self,
    const size_t buffer_position,
    const int64_t message_id, const uint32_t core_message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;

    /* buffer_position points after frame header and message header, padded */
    
    if ((buffer_position + body_size) <= MAX_OUTGOING_MESSAGE_PAYLOAD)
    {
        /* branch c1 */
        const size_t header_size = buffer_position - FRAME_HEADER_SIZE;
        const size_t frame_payload_size = yami_round_up_to_4(header_size + body_size);
        
        /* copy the body buffer */
        
        yami_memcpy(self->output_buffer_, buffer_position, body, 0u, body_size);
        
        /* and fill the frame header */
        
        yami_fill_frame_header(self->output_buffer_,
            core_message_id, header_size, frame_payload_size);
        
        self->output_buffer_has_data_ = 1; /* true, output buffer busy */
        self->output_data_size_ = FRAME_HEADER_SIZE + frame_payload_size;
        self->output_data_written_so_far_ = 0u;
        self->output_message_id_ = message_id;

        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        /* message is too big */
        
        res = yami_not_enough_space;
    }

    return res;
}

enum yami_result yami_channel_post_message(struct yami_channel * const self,
    const char object_name[], const char message_name[],
    const int64_t message_id, const uint32_t core_message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    if (self->output_buffer_has_data_ == 0)
    {
        /* branch c1 */
        /* serialize proper message header */

        size_t buffer_position = FRAME_HEADER_SIZE;
        
        res = yami_fill_message_header(
            self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
            &buffer_position,
            object_name, message_name, message_id);
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            res = yami_complete_msg_preparation(self, buffer_position,
                message_id, core_message_id, body, body_size);
        }
    }
    else
    {
        /* branch c2 */
        /* output buffer is busy */
        
        res = yami_not_enough_space;
    }
    
    return res;
}

enum yami_result yami_channel_post_reply(struct yami_channel * const self,
    const int64_t message_id, const uint32_t core_message_id,
    const uint8_t body[], const size_t body_size)
{
    enum yami_result res;
    
    if (self->output_buffer_has_data_ == 0)
    {
        /* branch c1 */
        /* serialize proper message header */

        size_t buffer_position = FRAME_HEADER_SIZE;
        
        yami_fill_reply_header(
            self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
            &buffer_position, message_id);

        res = yami_complete_msg_preparation(self, buffer_position,
            message_id, core_message_id, body, body_size);
    }
    else
    {
        /* branch c2 */
        /* output buffer is busy */
        
        res = yami_not_enough_space;
    }
    
    return res;
}

enum yami_result yami_channel_post_reject(struct yami_channel * const self,
    const int64_t message_id, const uint32_t core_message_id,
    const char reason[])
{
    enum yami_result res;
    
    if (self->output_buffer_has_data_ == 0)
    {
        /* branch c1 */
        /* serialize proper message header */

        size_t buffer_position = FRAME_HEADER_SIZE;
        
        res = yami_fill_exception_header(
            self->output_buffer_, MAX_OUTGOING_MESSAGE_PAYLOAD - FRAME_HEADER_SIZE,
            &buffer_position, reason, message_id);
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            /* the body buffer should be empty - that is, it contains 0 elements */
            
            const uint8_t empty_body[4] = { (uint8_t)0u, (uint8_t)0u, (uint8_t)0u, (uint8_t)0u };
            const size_t empty_body_size = sizeof(empty_body);

            res = yami_complete_msg_preparation(self, buffer_position,
                0LL, core_message_id, empty_body, empty_body_size);
        }
    }
    else
    {
        /* branch c2 */
        /* output buffer is busy */
        
        res = yami_not_enough_space;
    }
    
    return res;
}

static enum yami_result yami_channel_write_tcp_data(
    struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    size_t written;
    
    const enum yami_result res = yami_write_tcp(self->socket_,
        self->output_buffer_, self->output_data_written_so_far_, self->output_data_size_,
        &written,
        io_error_callback);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        self->output_data_written_so_far_ += written;
        
        if (self->output_data_written_so_far_ == self->output_data_size_)
        {
            /* branch c1_c1 */
            /* complete message has been pushed out             */
            /* mark the output buffer as ready for next message */
            
            self->output_buffer_has_data_ = 0; /* false */
        }
    }
    
    return res;
}

static enum yami_result yami_channel_write_udp_data(
    struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    const enum yami_result res = yami_send_udp(
        self->socket_, self->host_, self->port_,
        self->output_buffer_, self->output_data_size_,
        io_error_callback);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        /* complete message has been pushed out             */
        /* mark the output buffer as ready for next message */
            
        self->output_buffer_has_data_ = 0; /* false */
    }
    
    return res;
}

enum yami_result yami_channel_write_some_data(struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    if (self->protocol_ == yami_tcp)
    {
        /* branch c1 */
        res = yami_channel_write_tcp_data(self, io_error_callback);
    }
    else /* yami_udp */
    {
        /* branch c2 */
        res = yami_channel_write_udp_data(self, io_error_callback);
    }
    
    return res;
}

static enum yami_result yami_channel_check_header(
    struct yami_channel * const self,
    size_t * const out_expected_packet_size)
{
    enum yami_result res = yami_ok;

    /* first read message header, then payload */
    if (self->input_data_read_so_far_ < FRAME_HEADER_SIZE)
    {
        /* branch c1 */
        /* message header was not yet read */
        
        *out_expected_packet_size = FRAME_HEADER_SIZE;
    }
    else
    {
        /* branch c2 */
        /* message header was already read, expect more for frame payload */
        
        size_t position = 0u;
        size_t * const buffer_position = &position;
        
        res = yami_parse_frame_header(
            self->input_buffer_, FRAME_HEADER_SIZE, buffer_position,
            &self->input_message_id_, &self->input_message_header_size_,
            &self->input_message_payload_size_);
        
        if (res == yami_ok)
        {
            /* branch c2_c1 */
            /* prepare to read the rest of the frame */
            
            *out_expected_packet_size =
                FRAME_HEADER_SIZE + self->input_message_payload_size_;
        }
    }

    return res;
}

static enum yami_result yami_channel_check_body(
    struct yami_channel * const self,
    const size_t expected_packet_size)
{
    enum yami_result res = yami_unexpected_value;

    if (self->input_data_read_so_far_ == expected_packet_size)
    {
        /* branch c1 */
        if (expected_packet_size > FRAME_HEADER_SIZE)
        {
            /* branch c1_c1 */
            /* complete message has been read (frame with payload) */
            /* mark the input buffer as ready for consumption      */
                        
            self->input_buffer_has_space_ = 0; /* false */
            self->input_message_complete_ = 1; /* true */

            res = yami_ok;
        }
    }

    return res;
}

static enum yami_result yami_channel_read_tcp_data(
    struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    size_t expected_packet_size;

    enum yami_result res = yami_channel_check_header(
        self, &expected_packet_size);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        if (expected_packet_size <= MAX_INCOMING_MESSAGE_PAYLOAD)
        {
            /* branch c1_c1 */
            size_t readn;
    
            res = yami_read_tcp(self->socket_,
                self->input_buffer_, self->input_data_read_so_far_,
                expected_packet_size, &readn,
                io_error_callback);
            
            if (res == yami_ok)
            {
                /* branch c1_c1_c1 */
                self->input_data_read_so_far_ += readn;
                
                (void)yami_channel_check_body(self, expected_packet_size);
            }
        }
        else
        {
            /* branch c1_c2 */
            res = yami_not_enough_space;
        }
    }
    
    return res;
}

void yami_channel_clean_input_buffer(struct yami_channel * const self)
{
    self->input_buffer_has_space_ = 1; /* true */
    self->input_data_read_so_far_ = 0u;
    self->input_message_complete_ = 0; /* false */
}

enum yami_result yami_channel_check_udp_packet(
    struct yami_channel * const self)
{
    size_t expected_packet_size;

    enum yami_result res = yami_channel_check_header(
        self, &expected_packet_size);
    if (res == yami_ok)
    {
        /* branch c1 */
        res = yami_channel_check_body(self, expected_packet_size);
        if (res != yami_ok)
        {
            /* branch c1_c1 */
            /* if the packet is not complete, silently drop it */

            yami_channel_clean_input_buffer(self);
        }
    }

    return res;
}

static enum yami_result yami_channel_read_udp_data(
    struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    const int32_t format_target = 0; /* false */

    size_t readn;

    enum yami_result res = yami_receive_udp(self->socket_,
        self->target_, format_target,
        self->input_buffer_, MAX_INCOMING_MESSAGE_PAYLOAD, &readn,
        io_error_callback);

    if (res == yami_ok)
    {
        /* branch c1 */
        /* the UDP packet is read in its entirety, without */
        /* separate stages for header and body             */
        /* - verify both parts now                         */

        self->input_data_read_so_far_ = readn;

        res = yami_channel_check_udp_packet(self);
    }

    return res;
}

enum yami_result yami_channel_read_some_data(
    struct yami_channel * const self,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    
    if (self->protocol_ == yami_tcp)
    {
        /* branch c1 */
        res = yami_channel_read_tcp_data(self, io_error_callback);
    }
    else
    {
        /* branch c2 */
        res = yami_channel_read_udp_data(self, io_error_callback);
    }

    return res;
}
