/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../../agent.h"
#include "../../serialization.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* these variables are used */
/* to pass back answers from the server */
int32_t reply_arrived = 0;
int32_t sum;
int32_t difference;
int32_t product;
int32_t ratio_defined = 0;
int32_t ratio;

int64_t client_message_id;

void message_callback(const char target[],
    enum yami_message_type message_type,
    const char object_name[], const char message_name[], const char exception_msg[],
    int64_t message_id,
    const uint8_t body[], size_t body_size)
{
    enum yami_result res;
    
    int32_t num_of_entries;
    int32_t i;
    int32_t len;
    const char sum_tag[] = "sum";
    const char difference_tag[] = "difference";
    const char product_tag[] = "product";
    const char ratio_tag[] = "ratio";
    char field_name[11];            /* longest expected field name */
    const size_t longest_field_name_len = 11;
    enum yami_parameter_type type;
    
    /* buffer position is updated with each reading operation */
    size_t buffer_position = 0u;
    
    if (message_type != yami_message_reply)
    {
        /* ignore all other message types, we only expect regular replies */

        return;
    }
    
    /* this reply should match the message that was just sent to the server */
    
    assert(message_id == client_message_id);
    
    /* for simplicity this server does not allow unknown fields */
    
    res = yami_get_integer(body, body_size, &buffer_position, &num_of_entries);
    assert(res == yami_ok);

    for (i = 0; i != num_of_entries; ++i)
    {
        /* read length of field name */
    
        res = yami_get_integer(body, body_size, &buffer_position, &len);
        assert(res == yami_ok);
    
        if (len > longest_field_name_len)
        {
            return;
        }
        
        /* read the field name */
    
        res = yami_get_raw_cstring(body, body_size, &buffer_position,
            field_name, (size_t)len);
        assert(res == yami_ok);
    
        /* read type (expect integer for all fields) */
        
        res = yami_get_type(body, body_size, &buffer_position, &type);
        assert(res == yami_ok);
        
        if (type != yami_integer)
        {
            return;
        }
        
        if (strcmp(field_name, sum_tag) == 0)
        {
            /* read the "sum" integer value */
            
            res = yami_get_integer(body, body_size, &buffer_position, &sum);
            assert(res == yami_ok);
        }
        else if (strcmp(field_name, difference_tag) == 0)
        {
            /* read the "difference" integer value */
            
            res = yami_get_integer(body, body_size, &buffer_position, &difference);
            assert(res == yami_ok);
        }
        else if (strcmp(field_name, product_tag) == 0)
        {
            /* read the "product" integer value */
            
            res = yami_get_integer(body, body_size, &buffer_position, &product);
            assert(res == yami_ok);
        }
        else if (strcmp(field_name, ratio_tag) == 0)
        {
            /* read the "ratio" integer value */
            
            res = yami_get_integer(body, body_size, &buffer_position, &ratio);
            assert(res == yami_ok);
            
            ratio_defined = 1;
        }
    }
    
    reply_arrived = 1;
}

void serialize_message(uint8_t message_body[], size_t buffer_size,
    size_t * message_body_size, int32_t a, int32_t b)
{
    enum yami_result res;
    
    /* both arguments are packed in two integer fields named "a" and "b" */
    
    /* a message with two integer fields has: */
    /* - 2 as the number of fields            */
    /* - "a" as field name                    */
    /* - type of field                        */
    /* - field int value                      */
    /* - "b" as field name                    */
    /* - type of field                        */
    /* - field int value                      */
    
    /* message_body_size will be updated with each serialized element */
    /* and at the end will indicate the size of the whole message     */
    *message_body_size = 0u;

    /* number of fields */
    res = yami_put_integer(message_body, buffer_size, message_body_size, 2);
    assert(res == yami_ok);

    /* field name */
    res = yami_put_cstring(message_body, buffer_size, message_body_size, "a");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(message_body, buffer_size, message_body_size, yami_integer);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_integer(message_body, buffer_size, message_body_size, a);
    assert(res == yami_ok);

    /* field name */
    res = yami_put_cstring(message_body, buffer_size, message_body_size, "b");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(message_body, buffer_size, message_body_size, yami_integer);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_integer(message_body, buffer_size, message_body_size, b);
    assert(res == yami_ok);
}

void run_client(const char * server_address, int32_t a, int32_t b)
{
    enum yami_result res;
    
    struct yami_agent client_agent;

    uint8_t message_body[100]; /* serialized message body     */
    size_t message_body_size;   /* depending on content        */
    size_t timeout = 1000u;     /* example, 1 second           */

    /* create the client agent */
    /* and open a channel to the given server */

    yami_init(&client_agent);

    /* message callback will be used to receive replies from the server */
    
    yami_set_message_callback(&client_agent, &message_callback);

    res = yami_open_connection(&client_agent, server_address);
    assert(res == yami_ok);

    /* prepare the message body for sending to the server */
    
    serialize_message(message_body, sizeof(message_body),
        &message_body_size, a, b);

    /* post the message to outgoing channel */
    
    res = yami_post_message(&client_agent, server_address,
        "calculator", "calculate",
        message_body, message_body_size,
        &client_message_id);
    assert(res == yami_ok);
    
    /* do I/O work until the reply arrives */
    
    while (reply_arrived == 0)
    {
        res = yami_do_some_work(&client_agent, timeout);
        assert(res == yami_ok || res == yami_timed_out);
    }
    
    /* print results */

    printf("sum        = %d\n", sum);
    printf("difference = %d\n", difference);
    printf("product    = %d\n", product);

    printf("ratio      = ");
    if (ratio_defined == 1)
    {
        printf("%d", ratio);
    }
    else
    {
        printf("<undefined>");
    }

    puts("");

    yami_clean(&client_agent);
}

int main(int argc, char * argv[])
{
    const char * server_address;
    char * aendptr;
    char * bendptr;
    int32_t a;
    int32_t b;
    
    if (argc != 4)
    {
        puts("expecting three parameters: "
            "server destination and two integers");
        return EXIT_FAILURE;
    }

    server_address = argv[1];

    a = (int32_t)strtol(argv[2], &aendptr, 10);
    b = (int32_t)strtol(argv[3], &bendptr, 10);
    if ((*aendptr != '\0') || (*bendptr != '\0'))
    {
        puts("cannot parse the second or third parameter");
    }
    else
    {
        run_client(server_address, a, b);
    }
    
    return 0;
}
