/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../../agent.h"
#include "../../serialization.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

int32_t a;
int32_t b;
char client_target[100];
int64_t client_message_id;
int32_t request_arrived = 0;

void message_callback(const char target[],
    enum yami_message_type message_type,
    const char object_name[], const char message_name[], const char exception_msg[],
    int64_t message_id,
    const uint8_t body[], size_t body_size)
{
    /* ignore the header, but deserialize the message body */

    enum yami_result res;
    
    int32_t num_of_entries;
    int32_t i;
    int32_t len;
    char field_name[2]; /* expect only "a" and "b" */
    enum yami_parameter_type type;
    
    /* buffer position is updated with each reading operation */
    size_t buffer_position = 0u;
    
    /* remember client target so that we can send the reply back */
    
    strcpy(client_target, target);
    client_message_id = message_id;
    
    /* for simplicity this server does not allow unknown fields, */
    /* so the only expected message format is a regular message  */
    /* with "a" and "b" integer fields                           */
    
    res = yami_get_integer(body, body_size, &buffer_position, &num_of_entries);
    assert(res == yami_ok);
    
    if (num_of_entries != 2)
    {
        return;
    }
    
    for (i = 0; i != num_of_entries; ++i)
    {
        /* read length of field name */
        
        res = yami_get_integer(body, body_size, &buffer_position, &len);
        assert(res == yami_ok);
        
        if (len != 1)
        {
            return;
        }
        
        /* read the field name (expect "a" or "b") */
        
        res = yami_get_raw_cstring(body, body_size, &buffer_position,
            field_name, (size_t)len);
        assert(res == yami_ok);
        
        /* read type (expect integer for both fields) */
        
        res = yami_get_type(body, body_size, &buffer_position, &type);
        assert(res == yami_ok);
        
        if (type != yami_integer)
        {
            return;
        }
        
        if (strcmp(field_name, "a") == 0)
        {
            /* read "a" integer value */
                
            res = yami_get_integer(body, body_size, &buffer_position, &a);
            assert(res == yami_ok);
        }
        else
        {
            /* read "b" integer value */
                
            res = yami_get_integer(body, body_size, &buffer_position, &b);
            assert(res == yami_ok);
        }
    }
    
    request_arrived = 1;
}

void serialize_reply(uint8_t reply_body[], size_t buffer_size,
    size_t * reply_body_size)
{
    enum yami_result res;
    
    int32_t num_of_entries;
    
    /* compute the results and prepare reply body in appropriate format */
    
    /* note: if b == 0, the ratio field will not be present in the reply */
    
    if (b != 0)
    {
        num_of_entries = 4;
    }
    else
    {
        num_of_entries = 3;
    }
    
    /* message_body_size will be updated with each serialized element */
    /* and at the end will indicate the size of the whole message     */
    *reply_body_size = 0u;

    /* number of fields */
    res = yami_put_integer(reply_body, buffer_size, reply_body_size, num_of_entries);
    assert(res == yami_ok);

    /* put sum value */
    
    /* field name */
    res = yami_put_cstring(reply_body, buffer_size, reply_body_size, "sum");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(reply_body, buffer_size, reply_body_size, yami_integer);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_integer(reply_body, buffer_size, reply_body_size, a + b);
    assert(res == yami_ok);

    /* put difference value */
    
    /* field name */
    res = yami_put_cstring(reply_body, buffer_size, reply_body_size, "difference");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(reply_body, buffer_size, reply_body_size, yami_integer);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_integer(reply_body, buffer_size, reply_body_size, a - b);
    assert(res == yami_ok);

    /* put product value */
    
    /* field name */
    res = yami_put_cstring(reply_body, buffer_size, reply_body_size, "product");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(reply_body, buffer_size, reply_body_size, yami_integer);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_integer(reply_body, buffer_size, reply_body_size, a * b);
    assert(res == yami_ok);
    
    /* put ratio value, if possible */
    
    if (b != 0)
    {
        /* field name */
        res = yami_put_cstring(reply_body, buffer_size, reply_body_size, "ratio");
        assert(res == yami_ok);

        /* field type */
        res = yami_put_type(reply_body, buffer_size, reply_body_size, yami_integer);
        assert(res == yami_ok);
        
        /* field value */
        res = yami_put_integer(reply_body, buffer_size, reply_body_size, a / b);
        assert(res == yami_ok);
    }
}

void run_server(const char * server_address)
{
    enum yami_result res;

    uint8_t reply_body[100]; /* serialized message body     */
    size_t reply_body_size;   /* depending on content        */

    const size_t timeout = 1000; /* 1 second */
    
    struct yami_agent server_agent;

    /* create the server agent */
    /* and activate a listener on the given address */

    yami_init(&server_agent);
    
    /* message callback will be used to receive messages from client */
    
    yami_set_message_callback(&server_agent, &message_callback);

    res = yami_set_listener(&server_agent, server_address);
    assert(res == yami_ok);

    printf("The server is listening on %s\n", server_address);

    /* run the event loop */

    while (1)
    {
        res = yami_do_some_work(&server_agent, timeout);
        assert(res == yami_ok || res == yami_timed_out);
        
        if (request_arrived != 0)
        {
            /* request arrived and was already parsed,      */
            /* compute the result and prepare reply message */
            
            serialize_reply(reply_body, sizeof(reply_body),
                &reply_body_size);
            
            /* put the reply message in outgoing buffer */
            /* of the client's channel                  */
            
            res = yami_post_reply(&server_agent, client_target,
                client_message_id, reply_body, reply_body_size);
            assert(res == yami_ok);
            
            printf("got message with parameters %d and %d, "
                "response has been sent back\n", a, b);
        
            /* prepare for the next incoming message */
            
            request_arrived = 0;
        }
    }

    /* never reaches here */
}

int main(int argc, char * argv[])
{
    char * server_address;
    
    if (argc != 2)
    {
        puts("expecting one parameter: server destination");
        return EXIT_FAILURE;
    }

    server_address = argv[1];

    run_server(server_address);
    
    /* unreachable */
    return 0;
}
