/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "../../agent.h"
#include "../../serialization.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

void serialize_message(uint8_t message_body[], size_t buffer_size,
    size_t * message_body_size, const char input_line[])
{
    enum yami_result res;
    
    /* the "content" field name is arbitrary,        */
    /* but needs to be recognized at the server side */
    
    /* a message with single string field has: */
    /* - 1 as the number of fields             */
    /* - "content" as field name               */
    /* - type of field                         */
    /* - field string value                    */
    
    /* message_body_size will be updated with each serialized element */
    /* and at the end will indicate the size of the whole message     */
    *message_body_size = 0u;

    /* number of fields */
    res = yami_put_integer(message_body, buffer_size, message_body_size, 1);
    assert(res == yami_ok);

    /* field name */
    res = yami_put_cstring(message_body, buffer_size, message_body_size, "content");
    assert(res == yami_ok);

    /* field type */
    res = yami_put_type(message_body, buffer_size, message_body_size, yami_string);
    assert(res == yami_ok);
    
    /* field value */
    res = yami_put_cstring(message_body, buffer_size, message_body_size, input_line);
    assert(res == yami_ok);
}

void run_client(const char * server_address)
{
    enum yami_result res;
    
    struct yami_agent client_agent;

    char input_line[1000];      /* text as entered by the user */
    uint8_t message_body[1000]; /* serialized message body     */
    size_t message_body_size;   /* depending on content        */
    int64_t message_id;
    int32_t channel_is_busy;
    size_t timeout = 1000u;     /* example, 1 second           */

    /* create the client agent */
    /* and open a channel to the given server */

    yami_init(&client_agent);

    res = yami_open_connection(&client_agent, server_address);
    assert(res == yami_ok);

    /* read lines of text from standard input */
    /* and post each one for transmission */

    while (fgets(input_line, 1000, stdin))
    {
        /* remove newline character */
        
        input_line[strlen(input_line) - 1] = '\0';

        /* create message in proper YAMI4 format */
        
        serialize_message(message_body, sizeof(message_body),
            &message_body_size, input_line);

        /* post the message to outgoing channel         */
        /* "some_message" is arbitrary,                 */
        /* and is ignored by the server in this example */
        
        res = yami_post_message(&client_agent, server_address,
            "printer", "some_message",
            message_body, message_body_size,
            &message_id);
        assert(res == yami_ok);
        
        /* do I/O work until the message is pushed out */
        
        res = yami_is_output_channel_busy(&client_agent, server_address,
            &channel_is_busy);
        assert(res == yami_ok);
        while (channel_is_busy != 0)
        {
            res = yami_do_some_work(&client_agent, timeout);
            assert(res == yami_ok);

            res = yami_is_output_channel_busy(&client_agent, server_address,
                &channel_is_busy);
            assert(res == yami_ok);
        }
    }

    yami_clean(&client_agent);
}

int main(int argc, char * argv[])
{
    const char * server_address;
    
    if (argc != 2)
    {
        puts("expecting one parameter: server destination");
        return EXIT_FAILURE;
    }

    server_address = argv[1];

    run_client(server_address);
    
    return 0;
}

