/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "network_utils.h"

#include "limits.h"
#include "network_constants.h"
#include "utils.h"

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>


static const char tcp_prefix[] = "tcp://";
static const char udp_prefix[] = "udp://";
static const size_t len_of_prefix = 6u;

static enum yami_result yami_parse_host(const char hostname[],
    const size_t from, const size_t to, uint32_t * const out_host)
{
    enum yami_result res = yami_unexpected_value;
    size_t i = from;
    int32_t dot_index;
    int32_t converted = 0; /* false */
    size_t next_to;
    uint8_t a = (uint8_t)0u;
    uint8_t b = (uint8_t)0u;
    uint8_t c = (uint8_t)0u;
    uint8_t d = (uint8_t)0u;
    
    /* parse all four componentes of xxx.xxx.xxx.xxx */
    
    dot_index = yami_strfind(hostname, (char)'.', i, to);
    if (dot_index > 0)
    {
        /* branch c1 */
        next_to = (size_t)dot_index;
        converted = yami_string_to_uint8(hostname, i, next_to, &a);
        if (converted != 0)
        {
            /* branch c1_c1 */
            i = next_to + 1u;
        }
    }
    
    if (converted != 0)
    {
        /* branch c2 */
        dot_index = yami_strfind(hostname, (char)'.', i, to);
        if (dot_index > 0)
        {
            /* branch c2_c1 */
            next_to = (size_t)dot_index;
            converted = yami_string_to_uint8(hostname, i, next_to, &b);
            if (converted != 0)
            {
                /* branch c2_c1_c1 */
                i = next_to + 1u;
            }
        }
    }
    
    if (converted != 0)
    {
        /* branch c3 */
        dot_index = yami_strfind(hostname, (char)'.', i, to);
        if (dot_index > 0)
        {
            /* branch c3_c1 */
            next_to = (size_t)dot_index;
            converted = yami_string_to_uint8(hostname, i, next_to, &c);
            if (converted != 0)
            {
                /* branch c3_c1_c1 */
                i = next_to + 1u;
            }
        }
    }
    
    if (converted != 0)
    {
        /* branch c4 */
        converted = yami_string_to_uint8(hostname, i, to, &d);
    }
    
    if (converted != 0)
    {
        /* branch c5 */
        *out_host =
            ((uint32_t)a << 24) |
            ((uint32_t)b << 16) |
            ((uint32_t)c << 8) |
            (uint32_t)d;
        
        res = yami_ok;
    }
    
    if (res == yami_ok)
    {
        /* branch c6 */
        *out_host = htonl(*out_host);
    }

    return res;
}

static void yami_format_target(char out_target[],
    const enum yami_protocol protocol,
    const uint32_t host, const uint16_t port)
{
    /* this function never fails, MAX_LENGTH_OF_TARGET_NAME always has safe value */

    const uint32_t nhost = ntohl(host);
    const uint16_t nport = htons(port);

    const uint32_t a = (nhost & 0xff000000u) >> 24;
    const uint32_t b = (nhost & 0x00ff0000u) >> 16;
    const uint32_t c = (nhost & 0x0000ff00u) >> 8;
    const uint32_t d =  nhost & 0x000000ffu;

    int32_t last_index;

    if (protocol == yami_tcp)
    {
        /* branch c1 */
        yami_strncpy(out_target, tcp_prefix, len_of_prefix);
    }
    else /* yami_udp */
    {
        /* branch c2 */
        yami_strncpy(out_target, udp_prefix, len_of_prefix);
    }

    last_index = yami_uint32_to_string(a, out_target, len_of_prefix,
        MAX_LENGTH_OF_TARGET_NAME);

    ++last_index;
    out_target[(size_t)last_index] = (char)'.';
    ++last_index;
    
    last_index = yami_uint32_to_string(b, out_target, (size_t)last_index,
        MAX_LENGTH_OF_TARGET_NAME);
    
    ++last_index;
    out_target[(size_t)last_index] = (char)'.';
    ++last_index;
    
    last_index = yami_uint32_to_string(c, out_target, (size_t)last_index,
        MAX_LENGTH_OF_TARGET_NAME);
    
    ++last_index;
    out_target[(size_t)last_index] = (char)'.';
    ++last_index;
    
    last_index = yami_uint32_to_string(d, out_target, (size_t)last_index,
        MAX_LENGTH_OF_TARGET_NAME);
    
    ++last_index;
    out_target[(size_t)last_index] = (char)':';
    ++last_index;
    
    last_index = yami_uint32_to_string((uint32_t)nport, out_target, (size_t)last_index,
        MAX_LENGTH_OF_TARGET_NAME);
    
    ++last_index;
    out_target[(size_t)last_index] = (char)'\0';
}

static enum yami_result yami_parse_endpoint(const char target[],
    enum yami_protocol * const out_protocol,
    uint32_t * const out_host, uint16_t * const out_port)
{
    enum yami_result res;
    
    const size_t target_length = yami_strlen(target);

    if (yami_strncmp(target, tcp_prefix, len_of_prefix) == 0)
    {
        /* branch c1 */
        *out_protocol = yami_tcp;
        
        res = yami_ok;
    }
    else if (yami_strncmp(target, udp_prefix, len_of_prefix) == 0)
    {
        /* branch c2 */
        *out_protocol = yami_udp;
        
        res = yami_ok;
    }
    else
    {
        /* branch c3 */
        res = yami_bad_protocol;
    }

    if (res == yami_ok)
    {
        /* branch c4 */
        int32_t colon_position;
        
        res = yami_unexpected_value;
        
        colon_position = yami_strfind(target, (char)':', len_of_prefix, target_length);
        if (colon_position > 0)
        {
            /* branch c4_c1 */
            res = yami_parse_host(target, len_of_prefix, (size_t)colon_position, out_host);
            
            if (res == yami_ok)
            {
                /* branch c4_c1_c1 */
                const size_t ucolon_position = (size_t)colon_position;
                uint32_t uport;
                const int32_t port_converted = yami_string_to_uint32(target,
                    ucolon_position + 1u, target_length, &uport);
                if (port_converted != 0)
                {
                    /* branch c4_c1_c1_c1 */
                    *out_port = htons((uint16_t)uport);
                }
                else
                {
                    /* branch c4_c1_c1_c2 */
                    res = yami_unexpected_value;
                }
            }
        }
    }
    
    return res;
}

static enum yami_result yami_set_reuseaddr(const int32_t fd,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t value = 1;
    const int32_t cc = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR,
        &value, (socklen_t)sizeof(value));
    if (cc == 0)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot set socket option (SO_REUSEADDR).");
        }

        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_set_nonblocking(const int32_t fd,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t flags = fcntl(fd, F_GETFL, 0);
    if (flags != -1)
    {
        /* branch c1 */
        const uint32_t uflags = (uint32_t)flags;
        const uint32_t new_uflags = uflags | (uint32_t)O_NONBLOCK;

        const int32_t cc = fcntl(fd, F_SETFL, (int32_t)new_uflags);
        if (cc != -1)
        {
            /* branch c1_c1 */
            res = yami_ok;
        }
        else
        {
            /* branch c1_c2 */
            if (io_error_callback != NULL)
            {
                /* branch c1_c2_c1 */
                io_error_callback(errno,
                    "Cannot set non-block flag for outgoing connection.");
            }

            res = yami_io_error;
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno,
                "Cannot get file flags for outgoing connection.");
        }

        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_set_nodelay(const int32_t fd,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t value = 1;
    const int32_t cc = setsockopt(fd, IPPROTO_TCP, TCP_NODELAY,
        &value, (socklen_t)sizeof(value));
    if (cc == 0)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot set socket option (TCP_NODELAY).");
        }

        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_set_keepalive(const int32_t fd,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t value = 1;
    const int32_t cc = setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE,
        &value, (socklen_t)sizeof(value));
    if (cc == 0)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot set socket option (SO_KEEPALIVE).");
        }

        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_create_tcp_listener(
    const uint32_t host, const uint16_t port,
    const struct yami_options * const options, int32_t * const out_socket,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t SOCKET_ERROR = -1;
        
    const int32_t fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd != SOCKET_ERROR)
    {
        /* branch c1 */
        res = yami_ok;
            
        if (options->tcp_reuseaddr != 0)
        {
            /* branch c1_c1 */
            res = yami_set_reuseaddr(fd, io_error_callback);
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot create listener socket.");
        }

        res = yami_io_error;
    }

    if (res == yami_ok)
    {
        /* branch c3 */
        struct sockaddr_in local;
        int32_t cc;

        /* MISRA-C deviation, expected by POSIX. */
        yami_bzero((uint8_t *)&local, sizeof(local));

        local.sin_family = (uint16_t)AF_INET;
        local.sin_port = port;
        local.sin_addr.s_addr = host;

        /* MISRA-C deviation, expected by POSIX. */
        cc = bind(fd,
            (struct sockaddr *)(&local), (socklen_t)sizeof(local));

        if (cc != SOCKET_ERROR)
        {
            /* branch c3_c1 */
            const int32_t backlog = options->tcp_listen_backlog;

            cc = listen(fd, backlog);
            if (cc != SOCKET_ERROR)
            {
                /* branch c3_c1_c1 */
                *out_socket = fd;
            }
            else
            {
                /* branch c3_c1_c2 */
                (void)close(fd);
                
                if (io_error_callback != NULL)
                {
                    /* branch c3_c1_c2_c1 */
                    io_error_callback(errno, "Cannot listen on the listener socket.");
                }

                res = yami_io_error;
            }
        }
        else
        {
            /* branch c3_c2 */
            (void)close(fd);
                
            if (io_error_callback != NULL)
            {
                /* branch c3_c2_c1 */
                io_error_callback(errno, "Cannot bind the listener socket.");
            }

            res = yami_io_error;
        }
    }

    return res;
}

static enum yami_result yami_create_udp_listener(
    const uint32_t host, const uint16_t port,
    const struct yami_options * const options, int32_t * const out_socket,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res = yami_io_error;

    const int32_t SOCKET_ERROR = -1;
        
    const int32_t fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd != SOCKET_ERROR)
    {
        /* branch c1 */
        struct sockaddr_in local;
        int32_t cc;

        /* MISRA-C deviation, expected by POSIX. */
        yami_bzero((uint8_t *)&local, sizeof(local));

        local.sin_family = (uint16_t)AF_INET;
        local.sin_port = port;
        local.sin_addr.s_addr = host;

        /* MISRA-C deviation, expected by POSIX. */
        cc = bind(fd,
            (struct sockaddr *)(&local), (socklen_t)sizeof(local));

        if (cc != SOCKET_ERROR)
        {
            /* branch c1_c1 */
            *out_socket = fd;
            
            res = yami_ok;
        }
        else
        {
            /* branch c1_c2 */
            (void)close(fd);
                
            if (io_error_callback != NULL)
            {
                /* branch c1_c2_c1 */
                io_error_callback(errno, "Cannot bind the UDP socket.");
            }
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot create the listening socket.");
        }
    }
    
    (void)options; /* no UDP options in this version */

    return res;
}

enum yami_result yami_create_listener(const char target[],
    const struct yami_options * const options,
    int32_t * const out_socket,
    enum yami_protocol * const out_protocol,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    uint32_t host;
    uint16_t port;
    
    res = yami_parse_endpoint(target, out_protocol, &host, &port);
    if (res == yami_ok)
    {
        /* branch c1 */
        if (*out_protocol == yami_tcp)
        {
            /* branch c1_c1 */
            res = yami_create_tcp_listener(
                host, port, options, out_socket, io_error_callback);
        }
        else /* yami_udp */
        {
            /* branch c1_c2 */
            res = yami_create_udp_listener(
                host, port, options, out_socket, io_error_callback);
        }
    }
    
    return res;
}

enum yami_result yami_clean_listener(const int32_t sock,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    int32_t r;
    
    r = close(sock);
    if (r == 0)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot close listener socket.");
        }
        
        res = yami_io_error;
    }
    
    return res;
}

static enum yami_result yami_create_outgoing_tcp_conn(
    const uint32_t host, const uint16_t port,
    const struct yami_options * const options, int32_t * const out_socket,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t SOCKET_ERROR = -1;
        
    const int32_t fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd != SOCKET_ERROR)
    {
        /* branch c1 */
        struct sockaddr_in servaddr;
        int32_t cc;
        
        res = yami_ok;

        /* MISRA-C deviation, expected by POSIX. */
        yami_bzero((uint8_t *)&servaddr, sizeof(servaddr));
        
        servaddr.sin_family = (uint16_t)AF_INET;
        servaddr.sin_port = port;
        servaddr.sin_addr.s_addr = host;

        if (options->tcp_nonblocking != 0)
        {
            /* branch c1_c1 */
            res = yami_set_nonblocking(fd, io_error_callback);

            /* perform non-blocking connection */
            /* as part of non-blocking setting */
                
            if (res == yami_ok)
            {
                /* branch c1_c1_c1 */
                /* MISRA-C deviation, expected by POSIX. */
                cc = connect(fd,
                    (struct sockaddr *)&servaddr,
                    (socklen_t)sizeof(servaddr));
                
                if (cc < 0)
                {
                    /* branch c1_c1_c1_c1 */
                    if (errno == EINPROGRESS)
                    {
                        /* branch c1_c1_c1_c1_c1 */
                        /* wait with configured timeout          */
                        /* for completing the connection process */
                        /* timeout 0ms means wait until complete */
                        
                        struct timeval tv;
                        struct timeval * tv_to_use;
                        int32_t num_of_descriptors_to_test;
                        size_t timeout;
                        
                        fd_set write_set;
                        FD_ZERO(&write_set);
                        FD_SET(fd, &write_set);
                        num_of_descriptors_to_test = fd + 1;
                        
                        timeout = (size_t)options->tcp_connect_timeout;
                        
                        if (timeout > 0u)
                        {
                            /* branch c1_c1_c1_c1_c1_c1 */
                            const size_t timeout_sec = timeout / 1000u;
                            const size_t timeout_msec = timeout % 1000u;
                            const size_t timeout_usec = timeout_msec * 1000u;
                            
                            tv.tv_sec = (time_t)timeout_sec;
                            tv.tv_usec = (suseconds_t)timeout_usec;
                            tv_to_use = &tv;
                        }
                        else
                        {
                            /* branch c1_c1_c1_c1_c1_c2 */
                            tv_to_use = NULL;
                        }
                        
                        cc = select(num_of_descriptors_to_test,
                            NULL, &write_set, NULL, tv_to_use);
                        if (cc > 0)
                        {
                            /* branch c1_c1_c1_c1_c1_c3 */
                            /* socket was either connected in time */
                            /* or there was an error               */
                            
                            int32_t error = 0;
                            socklen_t error_len = (socklen_t)sizeof(error);
                            cc = getsockopt(fd,
                                SOL_SOCKET, SO_ERROR, &error, &error_len);
                            if (cc == 0)
                            {
                                /* branch c1_c1_c1_c1_c1_c3_c1 */
                                if (error != 0)
                                {
                                    /* branch c1_c1_c1_c1_c1_c3_c1_c1 */
                                    if (io_error_callback != NULL)
                                    {
                                        /* branch c1_c1_c1_c1_c1_c3_c1_c1_c1 */
                                        io_error_callback(error,
                                            "Cannot connect (non-blocking) the outgoing socket.");
                                    }
                                    
                                    res = yami_io_error;
                                }
                            }
                            else
                            {
                                /* branch c1_c1_c1_c1_c1_c3_c2 */
                                if (io_error_callback != NULL)
                                {
                                    /* branch c1_c1_c1_c1_c1_c3_c2_c1 */
                                    io_error_callback(errno,
                                        "Cannot connect (non-blocking) the outgoing socket, getsockopt failed.");
                                }
                                
                                res = yami_io_error;
                            }
                        }
                        else if (cc == 0)
                        {
                            /* branch c1_c1_c1_c1_c1_c4 */
                            res = yami_timed_out;
                        }
                        else
                        {
                            /* branch c1_c1_c1_c1_c1_c5 */
                            if (io_error_callback != NULL)
                            {
                                /* branch c1_c1_c1_c1_c1_c5_c1 */
                                io_error_callback(errno,
                                    "Cannot wait for non-blocking connect, select failed.");
                            }
                            
                            res = yami_io_error;
                        }
                    }
                    else
                    {
                        /* branch c1_c1_c1_c1_c2 */
                        if (io_error_callback != NULL)
                        {
                            /* branch c1_c1_c1_c1_c2_c1 */
                            io_error_callback(errno,
                                "Cannot connect (non-blocking) the outgoing socket.");
                        }
                        
                        res = yami_io_error;
                    }
                }
            }
        }
        else
        {
            /* branch c1_c2 */
            /* perform blocking connection */
            
            /* MISRA-C deviation, expected by POSIX. */
            cc = connect(fd,
                (struct sockaddr *)&servaddr,
                (socklen_t)sizeof(servaddr));

            if (cc == SOCKET_ERROR)
            {
                /* branch c1_c2_c1 */
                if (io_error_callback != NULL)
                {
                    /* branch c1_c2_c1_c1 */
                    io_error_callback(errno,
                        "Cannot connect the outgoing socket.");
                }

                res = yami_io_error;
            }
        }
            
        if (res == yami_ok)
        {
            /* branch c1_c3 */
            if (options->tcp_nodelay != 0)
            {
                /* branch c1_c3_c1 */
                res = yami_set_nodelay(fd, io_error_callback);
            }
        }
            
        if (res == yami_ok)
        {
            /* branch c1_c4 */
            if (options->tcp_keepalive != 0)
            {
                /* branch c1_c4_c1 */
                res = yami_set_keepalive(fd, io_error_callback);
            }
        }
            
        if (res == yami_ok)
        {
            /* branch c1_c5 */
            *out_socket = fd;
        }
        else
        {
            /* branch c1_c6 */
            (void)close(fd);
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot create outgoing TCP socket.");
        }

        res = yami_io_error;
    }

    return res;
}

static enum yami_result yami_create_outgoing_udp_conn(
    const struct yami_options * const options, int32_t * const out_socket,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    const int32_t SOCKET_ERROR = -1;
        
    const int32_t fd = socket(AF_INET, SOCK_DGRAM, 0);
    if (fd != SOCKET_ERROR)
    {
        /* branch c1 */
        *out_socket = fd;
        
        (void)options; /* no UDP options in this version */

        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot create outgoing UDP socket.");
        }

        res = yami_io_error;
    }

    return res;
}

enum yami_result yami_create_outgoing_connection(const char target[],
    const struct yami_options * const options,
    int32_t * const out_socket,
    enum yami_protocol * const out_protocol,
    uint32_t * const out_host,
    uint16_t * const out_port,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    
    res = yami_parse_endpoint(target, out_protocol, out_host, out_port);
    if (res == yami_ok)
    {
        /* branch c1 */
        if (*out_protocol == yami_tcp)
        {
            /* branch c1_c1 */
            res = yami_create_outgoing_tcp_conn(
                *out_host, *out_port, options, out_socket,
                io_error_callback);
        }
        else /* yami_udp */
        {
            /* branch c1_c2 */
            res = yami_create_outgoing_udp_conn(
                options, out_socket,
                io_error_callback);
        }
    }

    return res;
}

enum yami_result yami_accept_tcp(const int32_t listening_sock,
    const struct yami_options * const options, char out_target[], int32_t * const out_socket,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    struct sockaddr_in clientaddr;
    socklen_t clientaddr_len = (socklen_t)sizeof(clientaddr);
    
    const int32_t SOCKET_ERROR = -1;
    
    /* MISRA-C deviation, expected by POSIX. */
    const int32_t fd = accept(listening_sock,
        (struct sockaddr *)&clientaddr, &clientaddr_len);

    if (fd != SOCKET_ERROR)
    {
        /* branch c1 */
        res = yami_ok;
        
        yami_format_target(out_target, yami_tcp,
            clientaddr.sin_addr.s_addr,
            clientaddr.sin_port);
        
        if (options->tcp_nonblocking != 0)
        {
            /* branch c1_c1 */
            res = yami_set_nonblocking(fd, io_error_callback);
        }
        
        if (res == yami_ok)
        {
            /* branch c1_c2 */
            if (options->tcp_nodelay != 0)
            {
                /* branch c1_c2_c1 */
                res = yami_set_nodelay(fd, io_error_callback);
            }
        }
        
        if (res == yami_ok)
        {
            /* branch c1_c3 */
            if (options->tcp_keepalive != 0)
            {
                /* branch c1_c3_c1 */
                res = yami_set_keepalive(fd, io_error_callback);
            }
        }
        
        if (res == yami_ok)
        {
            /* branch c1_c4 */
            *out_socket = fd;
        }
        else
        {
            /* branch c1_c5 */
            (void)close(fd);
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot accept incoming TCP connection.");
        }
        
        res = yami_io_error;
    }

    return res;
}

enum yami_result yami_clean_connection(const int32_t sock,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    int32_t r;
    
    r = close(sock);
    if (r == 0)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot close connection.");
        }
        
        res = yami_io_error;
    }
    
    return res;
}

enum yami_result yami_write_tcp(const int32_t sock,
    const uint8_t data[], const size_t from, const size_t to, size_t * const out_written,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    
    const ssize_t wn = send(sock, &data[from], to - from, MSG_NOSIGNAL);
    if (wn == -1)
    {
        /* branch c1 */
        if (errno == EINTR)
        {
            /* branch c1_c1 */
            *out_written = 0u;
            
            res = yami_ok;
        }
        else
        {
            /* branch c1_c2 */
            if (io_error_callback != NULL)
            {
                /* branch c1_c2_c1 */
                io_error_callback(errno, "Cannot write to TCP socket.");
            }
            
            res = yami_io_error;
        }
    }
    else
    {
        /* branch c2 */
        *out_written = (size_t)(wn);
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_send_udp(const int32_t sock,
    const uint32_t host, const uint16_t port,
    const uint8_t data[], const size_t size,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;

    /* the UDP frame is either successfully pushed as a whole */
    /* or the operation is considered as failing              */

    struct sockaddr_in servaddr;
    ssize_t wn;

    /* MISRA-C deviation, expected by POSIX. */
    yami_bzero((uint8_t *)&servaddr, sizeof(servaddr));
        
    servaddr.sin_family = (uint16_t)AF_INET;
    servaddr.sin_port = port;
    servaddr.sin_addr.s_addr = host;

    /* MISRA-C deviation, expected by POSIX. */
    wn = sendto(sock, data, size, 0,
        (struct sockaddr *)&servaddr,
        (socklen_t)sizeof(servaddr));
    if ((size_t)wn == size)
    {
        /* branch c1 */
        res = yami_ok;
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot send UDP packet.");
        }
            
        res = yami_io_error;
    }

    return res;
}

enum yami_result yami_read_tcp(const int32_t sock,
    uint8_t out_data[], const size_t from, const size_t to,
    size_t * const out_readn,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    
    const ssize_t rn = read(sock, &out_data[from], to - from);
    if (rn == -1)
    {
        /* branch c1 */
        if (errno == EINTR)
        {
            /* branch c1_c1 */
            *out_readn = 0u;

            res = yami_ok;
        }
        else if (errno == ECONNRESET)
        {
            /* branch c1_c2 */
            res = yami_channel_closed;
        }
        else
        {
            /* branch c1_c3 */
            if (io_error_callback != NULL)
            {
                /* branch c1_c3_c1 */
                io_error_callback(errno, "Cannot read from socket.");
            }
            
            res = yami_io_error;
        }
    }
    else
    {
        /* branch c2 */
        if (rn != 0)
        {
            /* branch c2_c1 */
            *out_readn = (size_t)(rn);

            res = yami_ok;
        }
        else
        {
            /* branch c2_c2 */
            res = yami_channel_closed;
        }
    }
    
    return res;
}

enum yami_result yami_receive_udp(const int32_t sock,
    char out_target[], const int32_t format_target,
    uint8_t out_data[], const size_t size, size_t * const out_readn,
    const yami_io_error_function io_error_callback)
{
    /* the UDP frame has to be read as a single operation */

    enum yami_result res;
    
    struct sockaddr_in remote_address;
    socklen_t address_size = (socklen_t)sizeof(remote_address);

    /* MISRA-C deviation, expected by POSIX. */
    const ssize_t rn = recvfrom(sock, out_data, size, 0,
        (struct sockaddr *)&remote_address,
        &address_size);

    if (rn >= 0)
    {
        /* branch c1 */
        *out_readn = (size_t)rn;
        
        res = yami_ok;

        if (format_target != 0)
        {
            /* branch c1_c1 */
            /* compose the proper target name for this packet */

            yami_format_target(out_target, yami_udp,
                remote_address.sin_addr.s_addr,
                remote_address.sin_port);
        }
    }
    else
    {
        /* branch c2 */
        if (io_error_callback != NULL)
        {
            /* branch c2_c1 */
            io_error_callback(errno, "Cannot read UDP packet.");
        }
            
        res = yami_io_error;
    }
    
    return res;
}

enum yami_result yami_wait_for_work(
    struct yami_channel channels[], size_t channels_array_size,
    const int32_t listening_sock, int32_t * const out_listening_sock_ready,
    const size_t timeout,
    const yami_io_error_function io_error_callback)
{
    enum yami_result res;
    
    struct timeval tv;
    struct timeval * tv_to_use;
    int32_t num_of_descriptors_to_test = 0;
    int32_t has_some_free_channels = 0; /* false */
    int32_t cc;
    size_t i;
    
    fd_set read_set;
    fd_set write_set;

    FD_ZERO(&read_set);
    FD_ZERO(&write_set);
    
    for (i = 0u; i != channels_array_size; ++i)
    {
        /* branch 1 */
        const int32_t sock = channels[i].socket_;
        
        if (sock != INVALID_FILE_DESCRIPTOR)
        {
            /* branch 1_c1 */
            /* depending on availability of data or buffer space */
            /* put the channel's socket in the selector set      */
            
            if (channels[i].output_buffer_has_data_ != 0)
            {
                /* branch 1_c1_c1 */
                FD_SET(sock, &write_set);
                
                if (sock >= num_of_descriptors_to_test)
                {
                    /* branch 1_c1_c1_c1 */
                    num_of_descriptors_to_test = sock + 1;
                }
            }

            if (channels[i].input_buffer_has_space_ != 0)
            {
                /* branch 1_c1_c2 */
                FD_SET(sock, &read_set);
                
                if (sock >= num_of_descriptors_to_test)
                {
                    /* branch 1_c1_c2_c1 */
                    num_of_descriptors_to_test = sock + 1;
                }
            }
        }
        else
        {
            /* branch 1_c2 */
            has_some_free_channels = 1; /* true */
        }
    }
    
    /* decision has_place_to_check_listener: (has_some_free_channels != 0) && (listening_sock != INVALID_FILE_DESCRIPTOR) */
    /*                                    A: has_some_free_channels != 0               */
    /*                                    B: listening_sock != INVALID_FILE_DESCRIPTOR */
    /* end                                                                             */
    if ((has_some_free_channels != 0) &&
        (listening_sock != INVALID_FILE_DESCRIPTOR))
    {
        /* branch c2 */
        FD_SET(listening_sock, &read_set);
        
        if (listening_sock >= num_of_descriptors_to_test)
        {
            /* branch c2_c1 */
            num_of_descriptors_to_test = listening_sock + 1;
        }
    }

    if (num_of_descriptors_to_test > 0)
    {
        /* branch c3 */
        if (timeout > 0u)
        {
            /* branch c3_c1 */
            const size_t timeout_sec = timeout / 1000u;
            const size_t timeout_msec = timeout % 1000u;
            const size_t timeout_usec = timeout_msec * 1000u;
            
            tv.tv_sec = (time_t)timeout_sec;
            tv.tv_usec = (suseconds_t)timeout_usec;
            tv_to_use = &tv;
        }
        else
        {
            /* branch c3_c2 */
            tv_to_use = NULL;
        }
        
        cc = select(num_of_descriptors_to_test,
            &read_set, &write_set, NULL, tv_to_use);
        if (cc > 0)
        {
            /* branch c3_c3 */
            /* one of the channels is ready for work */

            for (i = 0u; i != channels_array_size; ++i)
            {
                /* branch c3_c3_1 */
                const int32_t sock = channels[i].socket_;
                
                if (sock != INVALID_FILE_DESCRIPTOR)
                {
                    /* branch c3_c3_1_c1 */
                    if (FD_ISSET(sock, &read_set))
                    {
                        /* branch c3_c3_1_c1_c1 */
                        channels[i].socket_ready_for_input_ = 1; /* true */
                    }
                    else
                    {
                        /* branch c3_c3_1_c1_c2 */
                        channels[i].socket_ready_for_input_ = 0; /* false */
                    }

                    if (FD_ISSET(sock, &write_set))
                    {
                        /* branch c3_c3_1_c1_c3 */
                        channels[i].socket_ready_for_output_ = 1; /* true */
                    }
                    else
                    {
                        /* branch c3_c3_1_c1_c4 */
                        channels[i].socket_ready_for_output_ = 0; /* false */
                    }
                }
            }
            
            /* check also the readiness of listening socket */
            
            if (listening_sock != INVALID_FILE_DESCRIPTOR)
            {
                /* branch c3_c3_c2 */
                if (FD_ISSET(listening_sock, &read_set))
                {
                    /* branch c3_c3_c2_c1 */
                    *out_listening_sock_ready = 1; /* true */
                }
                else
                {
                    /* branch c3_c3_c2_c2 */
                    *out_listening_sock_ready = 0; /* false */
                }
            }
            else
            {
                /* branch c3_c3_c3 */
                *out_listening_sock_ready = 0; /* false */
            }
            
            res = yami_ok;
        }
        else if (cc == 0)
        {
            /* branch c3_c4 */
            res = yami_timed_out;
        }
        else
        {
            /* branch c3_c5 */
            if (io_error_callback != NULL)
            {
                /* branch c3_c5_c1 */
                io_error_callback(errno,
                    "Cannot wait for work, select failed.");
            }

            res = yami_io_error;
        }
    }
    else
    {
        /* branch c4 */
        res = yami_bad_state;
    }
    
    return res;
}
