/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Note: this file implements serialization routines */
/* for both little-endian and big-endian platforms   */

#include "serialization.h"

#include "limits.h"
#include "network_constants.h"
#include "utils.h"

enum yami_result yami_put_type(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const enum yami_parameter_type type)
{
    enum yami_result res;
    
    switch (type)
    {
    case yami_boolean:
        /* branch c1 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 1);
        break;
    case yami_integer:
        /* branch c2 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 2);
        break;
    case yami_long_long:
        /* branch c3 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 3);
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float:
        /* branch c4 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 4);
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string:
        /* branch c5 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 5);
        break;
    case yami_binary:
        /* branch c6 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 6);
        break;
    case yami_boolean_array:
        /* branch c7 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 7);
        break;
    case yami_integer_array:
        /* branch c8 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 8);
        break;
    case yami_long_long_array:
        /* branch c9 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 9);
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float_array:
        /* branch c10 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 10);
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string_array:
        /* branch c11 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 11);
        break;
    case yami_binary_array:
        /* branch c12 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 12);
        break;
    case yami_nested_parameters:
        /* branch c13 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 13);
        break;
    case yami_nested_parameters_array:
        /* branch c14 */
        res = yami_put_integer(buffer, buffer_size, buffer_position, 14);
        break;
    default:
        res = yami_unexpected_value;
    }
    
    return res;
}

enum yami_result yami_get_type(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    enum yami_parameter_type * const type)
{
    enum yami_result res;
    int32_t code;
    
    res = yami_get_integer(buffer, buffer_size, buffer_position, &code);
    if (res == yami_ok)
    {
        /* branch c1 */
        switch (code)
        {
        case 1:
            /* branch c1_c1 */
            *type = yami_boolean;
            break;
        case 2:
            /* branch c1_c2 */
            *type = yami_integer;
            break;
        case 3:
            /* branch c1_c3 */
            *type = yami_long_long;
            break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
        case 4:
            /* branch c1_c4 */
            *type = yami_double_float;
            break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        case 5:
            /* branch c1_c5 */
            *type = yami_string;
            break;
        case 6:
            /* branch c1_c6 */
            *type = yami_binary;
            break;
        case 7:
            /* branch c1_c7 */
            *type = yami_boolean_array;
            break;
        case 8:
            /* branch c1_c8 */
            *type = yami_integer_array;
            break;
        case 9:
            /* branch c1_c9 */
            *type = yami_long_long_array;
            break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
        case 10:
            /* branch c1_c10 */
            *type = yami_double_float_array;
            break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        case 11:
            /* branch c1_c11 */
            *type = yami_string_array;
            break;
        case 12:
            /* branch c1_c12 */
            *type = yami_binary_array;
            break;
        case 13:
            /* branch c1_c13 */
            *type = yami_nested_parameters;
            break;
        case 14:
            /* branch c1_c14 */
            *type = yami_nested_parameters_array;
            break;
        default:
            /* branch c1_c15 */
            res = yami_unexpected_value;
        }
    }
    
    return res;
}

enum yami_result yami_put_integer(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 4u) /* sizeof(value) == 4 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        uint8_t byte0;
        uint8_t byte1;
        uint8_t byte2;
        uint8_t byte3;
        
        uint32_t uvalue;
        if (value >= 0)
        {
            /* branch c2_c1 */
            uvalue = (uint32_t)value;
        }
        else
        {
            /* branch c2_c2 */
            const int32_t tmp = -1 * (value + 1);
            uvalue = (uint32_t)tmp ^ 0xffffffffu;
        }
        
        byte0 = (uint8_t)(uvalue & 0xffu);
        byte1 = (uint8_t)((uvalue & 0xff00u) >> 8u);
        byte2 = (uint8_t)((uvalue & 0xff0000u) >> 16u);
        byte3 = (uint8_t)((uvalue & 0xff000000u) >> 24u);
        
        buffer[*buffer_position] = byte0;
        buffer[*buffer_position + 1u] = byte1;
        buffer[*buffer_position + 2u] = byte2;
        buffer[*buffer_position + 3u] = byte3;

        *buffer_position += 4u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_integer(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 4u) /* sizeof(*value) == 4 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const uint8_t byte0 = buffer[*buffer_position];
        const uint8_t byte1 = buffer[*buffer_position + 1u];
        const uint8_t byte2 = buffer[*buffer_position + 2u];
        const uint8_t byte3 = buffer[*buffer_position + 3u];
    
        const uint32_t uvalue =
            ((uint32_t)byte3 << 24) |
            ((uint32_t)byte2 << 16) |
            ((uint32_t)byte1 << 8) |
            (uint32_t)byte0;
        
        if ((uvalue & 0x80000000u) == 0u)
        {
            /* branch c2_c1 */
            /* positive number */
            
            *value = (int32_t)uvalue;
        }
        else
        {
            /* branch c2_c2 */
            /* negative number */
            
            const uint32_t temp = uvalue ^ 0xffffffffu;
            *value = -1 * (int32_t)temp - 1;
        }
        
        *buffer_position += 4u;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_long_long(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(value) == 8 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        uint8_t byte0;
        uint8_t byte1;
        uint8_t byte2;
        uint8_t byte3;
        uint8_t byte4;
        uint8_t byte5;
        uint8_t byte6;
        uint8_t byte7;
        
        uint64_t uvalue;
        if (value >= 0LL)
        {
            /* branch c2_c1 */
            uvalue = (uint64_t)value;
        }
        else
        {
            /* branch c2_c2 */
            const int64_t tmp = -1LL * (value + 1LL);
            uvalue = (uint64_t)tmp ^ 0xffffffffffffffffull;
        }
        
        byte0 = (uint8_t)(uvalue & 0xffull);
        byte1 = (uint8_t)((uvalue & 0xff00ull) >> 8u);
        byte2 = (uint8_t)((uvalue & 0xff0000ull) >> 16u);
        byte3 = (uint8_t)((uvalue & 0xff000000ull) >> 24u);
        byte4 = (uint8_t)((uvalue & 0xff00000000ull) >> 32u);
        byte5 = (uint8_t)((uvalue & 0xff0000000000ull) >> 40u);
        byte6 = (uint8_t)((uvalue & 0xff000000000000ull) >> 48u);
        byte7 = (uint8_t)((uvalue & 0xff00000000000000ull) >> 56u);
        
        buffer[*buffer_position] = byte0;
        buffer[*buffer_position + 1u] = byte1;
        buffer[*buffer_position + 2u] = byte2;
        buffer[*buffer_position + 3u] = byte3;
        buffer[*buffer_position + 4u] = byte4;
        buffer[*buffer_position + 5u] = byte5;
        buffer[*buffer_position + 6u] = byte6;
        buffer[*buffer_position + 7u] = byte7;

        *buffer_position += 8u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_long_long(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int64_t * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(*value) == 8 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const uint8_t byte0 = buffer[*buffer_position];
        const uint8_t byte1 = buffer[*buffer_position + 1u];
        const uint8_t byte2 = buffer[*buffer_position + 2u];
        const uint8_t byte3 = buffer[*buffer_position + 3u];
        const uint8_t byte4 = buffer[*buffer_position + 4u];
        const uint8_t byte5 = buffer[*buffer_position + 5u];
        const uint8_t byte6 = buffer[*buffer_position + 6u];
        const uint8_t byte7 = buffer[*buffer_position + 7u];

        const uint64_t uvalue =
            ((uint64_t)byte7 << 56) |
            ((uint64_t)byte6 << 48) |
            ((uint64_t)byte5 << 40) |
            ((uint64_t)byte4 << 32) |
            ((uint64_t)byte3 << 24) |
            ((uint64_t)byte2 << 16) |
            ((uint64_t)byte1 << 8) |
            (uint64_t)byte0;
        
        if ((uvalue & 0x8000000000000000ull) == 0ull)
        {
            /* branch c2_c1 */
            /* positive number */
            
            *value = (int64_t)uvalue;
        }
        else
        {
            /* branch c2_c2 */
            /* negative number */
            
            const uint64_t temp = uvalue ^ 0xffffffffffffffffull;
            *value = -1LL * (int64_t)temp - 1LL;
        }
        
        *buffer_position += 8u;

        res = yami_ok;
    }
    
    return res;
}

#ifdef YAMI4_WITH_DOUBLE_FLOAT

enum yami_result yami_put_double_float(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const double value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(value) == 8 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const uint8_t * p = (const uint8_t *)&value;
        
        buffer[*buffer_position] = *p++;
        buffer[*buffer_position + 1u] = *p++;
        buffer[*buffer_position + 2u] = *p++;
        buffer[*buffer_position + 3u] = *p++;
        buffer[*buffer_position + 4u] = *p++;
        buffer[*buffer_position + 5u] = *p++;
        buffer[*buffer_position + 6u] = *p++;
        buffer[*buffer_position + 7u] = *p;

        *buffer_position += 8u;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_double_float(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    double * const value)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < 8u) /* sizeof(*value) == 8 */
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        uint8_t * p = (uint8_t *)value;
        
        *p++ = buffer[*buffer_position];
        *p++ = buffer[*buffer_position + 1u];
        *p++ = buffer[*buffer_position + 2u];
        *p++ = buffer[*buffer_position + 3u];
        *p++ = buffer[*buffer_position + 4u];
        *p++ = buffer[*buffer_position + 5u];
        *p++ = buffer[*buffer_position + 6u];
        *p   = buffer[*buffer_position + 7u];
        
        *buffer_position += 8u;

        res = yami_ok;
    }
    
    return res;
}

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

enum yami_result yami_put_cstring(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char value[])
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* strings are prefixed with length */

    const size_t value_length = yami_strlen(value);
    
    /* length prefix with string length */
    const size_t space_needed = yami_round_up_to_4(4u + value_length);
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;
        size_t i;
        
        /* note: yami_put_integer might fail with small buffer, */
        /* but this was already checked above                   */
        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            buffer[*buffer_position + i] = (uint8_t)value[i];
        }
        
        /* add padding to align at the word boundary if necessary */
        
        while ((i + 4u) < space_needed)
        {
            /* branch c2_2 */
            buffer[*buffer_position + i] = (uint8_t)0u;
            ++i;
        }
        
        *buffer_position += i;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_cstring(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    char value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    const size_t value_length_rounded = yami_round_up_to_4(value_length);
    
    enum yami_result res;

    if (value_length_rounded > space_left)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i;
        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            value[i] = (char)buffer[*buffer_position + i];
        }
        
        value[i] = (char)'\0';
        
        *buffer_position += value_length_rounded;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_binary(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const uint8_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* binaries are prefixed with length */

    /* length prefix with binary length */
    const size_t space_needed = yami_round_up_to_4(4u + value_length);
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;
        size_t i;
        
        /* note: yami_put_integer might fail with small buffer, */
        /* but this was already checked above                   */
        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            buffer[*buffer_position + i] = value[i];
        }
        
        /* add padding to align at the word boundary if necessary */
        
        while ((i + 4u) < space_needed)
        {
            /* branch c2_2 */
            buffer[*buffer_position + i] = (uint8_t)0u;
            ++i;
        }
        
        *buffer_position += i;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_binary(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    uint8_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    const size_t value_length_rounded = yami_round_up_to_4(value_length);
    
    enum yami_result res;

    if (value_length_rounded > space_left)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i;
        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            value[i] = buffer[*buffer_position + i];
        }
        
        *buffer_position += value_length_rounded;
        
        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_boolean_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* boolean arrays are packed and are prefixed with length */

    const size_t length_padded = yami_round_up_to_8(value_length);
    const size_t space_needed_for_array = yami_round_up_to_4(length_padded / 8u);
    const size_t total_space_needed = 4u + space_needed_for_array;
    
    if (space_left < total_space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;

        size_t i = 0u; /* buffer byte index */
        size_t j = 0u; /* buffer bit index  */
        size_t k = 0u; /* value array index */
        
        /* note: yami_put_integer might fail with small buffer, */
        /* but this was already checked above                   */
        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        buffer[*buffer_position] = (uint8_t)0u;
        while (k < value_length)
        {
            /* branch c2_1 */
            if (value[k] != 0)
            {
                /* branch c2_1_c1 */
                const uint32_t mask = (uint32_t)1u << j;
                const uint32_t new_value = (uint32_t)buffer[*buffer_position + i] | mask;
                buffer[*buffer_position + i] = (uint8_t)new_value;
            }
            
            ++k;
            ++j;
            if (j == 8u)
            {
                /* branch c2_1_c2 */
                j = 0u;
                ++i;
                if (k < value_length)
                {
                    /* branch c2_1_c2_c1 */
                    buffer[*buffer_position + i] = (uint8_t)0u;
                }
            }
        }
        
        if (j != 0u)
        {
            /* branch c2_c2 */
            /* the packed array finished in the middle of last byte */
            ++i;
        }
        
        while (i < space_needed_for_array)
        {
            /* branch c2_3 */
            buffer[*buffer_position + i] = (uint8_t)0u;
            ++i;
        }
        
        *buffer_position += space_needed_for_array;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_boolean_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* boolean arrays are packed */

    const size_t length_padded = yami_round_up_to_8(value_length);
    const size_t space_needed = yami_round_up_to_4(length_padded / 8u);
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i = 0u; /* buffer byte index */
        size_t j = 0u; /* buffer bit index  */
        size_t k = 0u; /* value array index */
        
        while (k < value_length)
        {
            /* branch c2_1 */
            const uint32_t mask = (uint32_t)1u << j;
            const uint32_t b = (uint32_t)buffer[*buffer_position + i];
            
            if ((b & mask) != 0u)
            {
                /* branch c2_1_c1 */
                value[k] = 1;
            }
            else
            {
                /* branch c2_1_c2 */
                value[k] = 0;
            }

            ++k;
            ++j;
            if (j == 8u)
            {
                /* branch c2_1_c3 */
                j = 0u;
                ++i;
            }
        }
        
        *buffer_position += space_needed;

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_integer_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + (4u * value_length);
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;
        size_t i;

        /* the only way for yami_put_integer to fail is with not enough buffer space */
        /* - but that condition was already checked above                            */

        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            (void) yami_put_integer(buffer, buffer_size, buffer_position, value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_integer_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int32_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 4u * value_length;
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i;

        /* the only way for yami_get_integer to fail is with not enough buffer space */
        /* - but that condition was already checked above                            */

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            (void) yami_get_integer(buffer, buffer_size, buffer_position, &value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_put_long_long_array(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + (8u * value_length);
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;
        size_t i;
        
        /* the only way for yami_put_integer and yami_put_long_long to fail               */
        /* is with not enough buffer space - but that condition was already checked above */

        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            (void) yami_put_long_long(buffer, buffer_size, buffer_position, value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_long_long_array(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    int64_t value[], const size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 8u * value_length;
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i;

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            /* the only way for yami_get_long_long to fail is with not enough buffer space */
            /* - but that condition was already checked above                              */

            (void) yami_get_long_long(buffer, buffer_size, buffer_position, &value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

#ifdef YAMI4_WITH_DOUBLE_FLOAT

enum yami_result yami_put_double_float_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const double value[], size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    /* arrays are prefixed with length */

    /* length prefix with array length */
    const size_t space_needed = 4u + 8u * value_length;
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        const int32_t len = (int32_t)value_length;
        size_t i;
        
        /* the only way for yami_put_integer and yami_put_double_float_array to fail      */
        /* is with not enough buffer space - but that condition was already checked above */

        (void) yami_put_integer(buffer, buffer_size, buffer_position, len);

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            (void) yami_put_double_float(buffer, buffer_size, buffer_position, value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

enum yami_result yami_get_raw_double_float_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value[], size_t value_length)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    const size_t space_needed = 8u * value_length;
    
    if (space_left < space_needed)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        size_t i;

        /* the only way for yami_get_double_float to fail is with not enough buffer space */
        /* - but that condition was already checked above                                 */

        for (i = 0u; i != value_length; ++i)
        {
            /* branch c2_1 */
            (void) yami_get_double_float(buffer, buffer_size, buffer_position, &value[i]);
        }

        res = yami_ok;
    }
    
    return res;
}

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

enum yami_result yami_skip_field(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const enum yami_parameter_type type)
{
    enum yami_result res;
    
    switch (type)
    {
    case yami_boolean:
        /* branch c1 */
    case yami_integer:
        /* branch c2 */
        *buffer_position += 4u;
        res = yami_ok;
        break;
    case yami_long_long:
        /* branch c3 */
        *buffer_position += 8u;
        res = yami_ok;
        break;
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float:
        /* branch c4 */
        *buffer_position += 8u;
        res = yami_ok;
        break;
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
    case yami_string:
        /* branch c5 */
    case yami_binary:
        /* branch c6 */
        {
            /* branch c6_1 */
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                /* branch c6_1_c1 */
                *buffer_position += yami_round_up_to_4((size_t)len);
            }
        }
        break;
    case yami_boolean_array:
        /* branch c7 */
        {
            /* branch c7_1 */
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                /* branch c7_1_c1 */
                const size_t length_padded = yami_round_up_to_8((size_t)len);
                const size_t space_taken = yami_round_up_to_4(length_padded / 8u);

                *buffer_position += space_taken;
            }
        }
        break;
    case yami_integer_array:
        /* branch c8 */
        {
            /* branch c8_1 */
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                /* branch c8_1_c1 */
                *buffer_position += 4u * (size_t)len;
            }
        }
        break;
    case yami_long_long_array:
        /* branch c9 */
#ifdef YAMI4_WITH_DOUBLE_FLOAT
    case yami_double_float_array:
        /* branch c10 */
#endif /* YAMI4_WITH_DOUBLE_FLOAT */
        {
            /* branch c10_1 */
            int32_t len;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
            if (res == yami_ok)
            {
                /* branch c10_1_c1 */
                *buffer_position += 8u * (size_t)len;
            }
        }
        break;
    case yami_string_array:
        /* branch c11 */
    case yami_binary_array:
        /* branch c12 */
        {
            /* branch c12_1 */
            int32_t array_len;
            size_t i;
            res = yami_get_integer(buffer, buffer_size, buffer_position, &array_len);
            if (res == yami_ok)
            {
                /* branch c12_1_c1 */
                for (i = 0u; i != (size_t)array_len; ++i)
                {
                    /* branch c12_1_c1_1 */
                    int32_t string_len;
                    res = yami_get_integer(buffer, buffer_size, buffer_position, &string_len);
                    if (res == yami_ok)
                    {
                        /* branch c12_1_c1_1_c1 */
                        *buffer_position += yami_round_up_to_4((size_t)string_len);
                    }
                    else
                    {
                        /* branch c12_1_c1_1_c2 */
                        break;
                    }
                }
            }
        }
        break;
    case yami_nested_parameters:
        /* branch c13 */
    case yami_nested_parameters_array:
        /* branch c14 */
    default:
        res = yami_unexpected_value;
    }
    
    return res;
}

enum yami_result yami_fill_message_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char object_name[], const char message_name[], const int64_t message_id)
{
    enum yami_result res;

    /* message header contains 4 fields: */
    /* - message type                    */
    /* - object name                     */
    /* - message name                    */
    /* - message id                      */
    
    res = yami_put_integer(buffer, buffer_size, buffer_position, 4);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        const char type_tag[] = "type";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            /* branch c1_c2 */
            const char message_type[] = "message";
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_type);
        }
    }

    if (res == yami_ok)
    {
        /* branch c2 */
        const char object_name_tag[] = "object_name";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, object_name_tag);

        if (res == yami_ok)
        {
            /* branch c2_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            /* branch c2_c2 */
            res = yami_put_cstring(buffer, buffer_size, buffer_position, object_name);
        }
    }
    
    if (res == yami_ok)
    {
        /* branch c3 */
        const char msg_name_tag[] = "message_name";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_name_tag);

        if (res == yami_ok)
        {
            /* branch c3_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            /* branch c3_c2 */
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_name);
        }
    }

    if (res == yami_ok)
    {
        /* branch c4 */
        const char msg_id_tag[] = "message_id";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

        if (res == yami_ok)
        {
            /* branch c4_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);
        }

        if (res == yami_ok)
        {
            /* branch c4_c2 */
            res = yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
        }
    }

    return res;
}

void yami_fill_reply_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const int64_t message_id)
{
    /* this function assumes the buffer is big enough */
    /* and does not check for buffer size             */

    /* reply header contains 2 fields: */
    /* - message type                  */
    /* - message id                    */
    
    const char type_tag[] = "type";
    const char message_type[] = "reply";
    const char msg_id_tag[] = "message_id";

    (void) yami_put_integer(buffer, buffer_size, buffer_position, 2);
    
    (void) yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
    (void) yami_put_type(buffer, buffer_size, buffer_position, yami_string);

    (void) yami_put_cstring(buffer, buffer_size, buffer_position, message_type);

    (void) yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

    (void) yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);

    (void) yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
}

enum yami_result yami_fill_exception_header(uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    const char reason[], const int64_t message_id)
{
    enum yami_result res;

    /* exception header contains 3 fields: */
    /* - message type                      */
    /* - message id                        */
    /* - exception reason                  */
    
    res = yami_put_integer(buffer, buffer_size, buffer_position, 3);
    
    if (res == yami_ok)
    {
        /* branch c1 */
        const char type_tag[] = "type";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, type_tag);
        
        if (res == yami_ok)
        {
            /* branch c1_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            /* branch c1_c2 */
            const char message_type[] = "exception";
            res = yami_put_cstring(buffer, buffer_size, buffer_position, message_type);
        }
    }

    if (res == yami_ok)
    {
        /* branch c2 */
        const char msg_id_tag[] = "message_id";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, msg_id_tag);

        if (res == yami_ok)
        {
            /* branch c2_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_long_long);
        }

        if (res == yami_ok)
        {
            /* branch c2_c2 */
            res = yami_put_long_long(buffer, buffer_size, buffer_position, message_id);
        }
    }

    if (res == yami_ok)
    {
        /* branch c3 */
        const char reason_tag[] = "reason";
        res = yami_put_cstring(buffer, buffer_size, buffer_position, reason_tag);

        if (res == yami_ok)
        {
            /* branch c3_c1 */
            res = yami_put_type(buffer, buffer_size, buffer_position, yami_string);
        }

        if (res == yami_ok)
        {
            /* branch c3_c2 */
            res = yami_put_cstring(buffer, buffer_size, buffer_position, reason);
        }
    }

    return res;
}

enum yami_result yami_parse_message_header(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    char out_header_type[], const size_t header_type_size,
    char out_object_name[], const size_t object_name_size,
    char out_message_name[], const size_t message_name_size,
    char out_reason[], const size_t reason_size,
    int64_t * const out_message_id)
{
    enum yami_result res;
    int32_t tmp;
    
    const char type_tag[] = "type";
    const char object_name_tag[] = "object_name";
    const char message_name_tag[] = "message_name";
    const char message_id_tag[] = "message_id";
    const char reason_tag[] = "reason";
    
    /* set the output parameters to "safe" values */

    out_header_type[0] = (char)'\0';
    out_object_name[0] = (char)'\0';
    out_message_name[0] = (char)'\0';
    out_reason[0] = (char)'\0';
    *out_message_id = 0LL;
    
    res = yami_get_integer(buffer, buffer_size, buffer_position, &tmp);
    if (res == yami_ok)
    {
        /* branch c1 */
        if (tmp > 0)
        {
            /* branch c1_c1 */
            const size_t num_of_entries = (size_t)tmp;
            size_t i;
            int32_t len;
            char name[MAX_HEADER_TYPE_LENGTH]; /* 13, the length of longest expected field name */
            enum yami_parameter_type type;
            
            for (i = 0u; i != num_of_entries; ++i)
            {
                /* branch c1_c1_1 */
                /* read name of field */
                
                res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                if (res == yami_ok)
                {
                    /* branch c1_c1_1_c1 */
                    if (len > 0)
                    {
                        /* branch c1_c1_1_c1_c1 */
                        if ((size_t)len < MAX_HEADER_TYPE_LENGTH)
                        {
                            /* branch c1_c1_1_c1_c1_c1 */
                            res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                name, (size_t)len);
                        }
                        else
                        {
                            /* branch c1_c1_1_c1_c1_c2 */
                            res = yami_unexpected_value;
                        }
                    }
                    else
                    {
                        /* branch c1_c1_1_c1_c2 */
                        res = yami_unexpected_value;
                    }
                }
                
                /* read type of field */
                
                if (res == yami_ok)
                {
                    /* branch c1_c1_1_c2 */
                    res = yami_get_type(buffer, buffer_size, buffer_position, &type);
                }
                
                /* read field values according to their names and types */
                
                if (res == yami_ok)
                {
                    /* branch c1_c1_1_c3 */
                    if (type == yami_string)
                    {
                        /* branch c1_c1_1_c3_c1 */
                        /* this might be one of string fields that describe the message header: */
                        /* "type", "object_name", "message_name", "reason" or some unknown,     */
                        /* but legal (optional) string field                                    */

                        if (yami_strcmp(name, type_tag) == 0)
                        {
                            /* branch c1_c1_1_c3_c1_c1 */
                            /* this is the message "type" field */
                            
                            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                            if (res == yami_ok)
                            {
                                /* branch c1_c1_1_c3_c1_c1_c1 */
                                if (len > 0)
                                {
                                    /* branch c1_c1_1_c3_c1_c1_c1_c1 */
                                    if ((size_t)len < header_type_size)
                                    {
                                        /* branch c1_c1_1_c3_c1_c1_c1_c1_c1 */
                                        res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                            out_header_type, (size_t)len);
                                    }
                                    else
                                    {
                                        /* branch c1_c1_1_c3_c1_c1_c1_c1_c2 */
                                        res = yami_not_enough_space;
                                    }
                                }
                                else
                                {
                                    /* branch c1_c1_1_c3_c1_c1_c1_c2 */
                                    res = yami_unexpected_value;
                                }
                            }
                        }
                        else if (yami_strcmp(name, object_name_tag) == 0)
                        {
                            /* branch c1_c1_1_c3_c1_c2 */
                            /* this is the "object_name" field */
                            
                            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                            if (res == yami_ok)
                            {
                                /* branch c1_c1_1_c3_c1_c2_c1 */
                                if (len >= 0)
                                {
                                    /* branch c1_c1_1_c3_c1_c2_c1_c1 */
                                    if ((size_t)len < object_name_size)
                                    {
                                        /* branch c1_c1_1_c3_c1_c2_c1_c1_c1 */
                                        res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                            out_object_name, (size_t)len);
                                    }
                                    else
                                    {
                                        /* branch c1_c1_1_c3_c1_c2_c1_c1_c2 */
                                        res = yami_not_enough_space;
                                    }
                                }
                                else
                                {
                                    /* branch c1_c1_1_c3_c1_c2_c1_c2 */
                                    res = yami_unexpected_value;
                                }
                            }
                        }
                        else if (yami_strcmp(name, message_name_tag) == 0)
                        {
                            /* branch c1_c1_1_c3_c1_c3 */
                            /* this is "message_name" field */
                            
                            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                            if (res == yami_ok)
                            {
                                /* branch c1_c1_1_c3_c1_c3_c1 */
                                if (len >= 0)
                                {
                                    /* branch c1_c1_1_c3_c1_c3_c1_c1 */
                                    if ((size_t)len < message_name_size)
                                    {
                                        /* branch c1_c1_1_c3_c1_c3_c1_c1_c1 */
                                        res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                            out_message_name, (size_t)len);
                                    }
                                    else
                                    {
                                        /* branch c1_c1_1_c3_c1_c3_c1_c1_c2 */
                                        res = yami_not_enough_space;
                                    }
                                }
                                else
                                {
                                    /* branch c1_c1_1_c3_c1_c3_c1_c2 */
                                    res = yami_unexpected_value;
                                }
                            }
                        }
                        else if (yami_strcmp(name, reason_tag) == 0)
                        {
                            /* branch c1_c1_1_c3_c1_c4 */
                            /* this is "reason" field */
                            
                            res = yami_get_integer(buffer, buffer_size, buffer_position, &len);
                            if (res == yami_ok)
                            {
                                /* branch c1_c1_1_c3_c1_c4_c1 */
                                if (len >= 0)
                                {
                                    /* branch c1_c1_1_c3_c1_c4_c1_c1 */
                                    if ((size_t)len < reason_size)
                                    {
                                        /* branch c1_c1_1_c3_c1_c4_c1_c1_c1 */
                                        res = yami_get_raw_cstring(buffer, buffer_size, buffer_position,
                                            out_reason, (size_t)len);
                                    }
                                    else
                                    {
                                        /* branch c1_c1_1_c3_c1_c4_c1_c1_c2 */
                                        res = yami_not_enough_space;
                                    }
                                }
                                else
                                {
                                    /* branch c1_c1_1_c3_c1_c4_c1_c2 */
                                    res = yami_unexpected_value;
                                }
                            }
                        }
                        else
                        {
                            /* branch c1_c1_1_c3_c1_c5 */
                            /* unknown, but legal (optional) string field */
                            
                            res = yami_skip_field(buffer, buffer_size, buffer_position, type);
                        }
                    }
                    else if (type == yami_long_long)
                    {
                        /* branch c1_c1_1_c3_c2 */
                        /* this might be the "message_id" field that is part of message header */

                        if (yami_strcmp(name, message_id_tag) == 0)
                        {
                            /* branch c1_c1_1_c3_c2_c1 */
                            /* this is "message_id" field */
                            
                            res = yami_get_long_long(buffer, buffer_size, buffer_position, out_message_id);
                        }
                        else
                        {
                            /* branch c1_c1_1_c3_c2_c2 */
                            /* unknown, but legal (optional) long-long field */
                            
                            res = yami_skip_field(buffer, buffer_size, buffer_position, type);
                        }
                    }
                    else
                    {
                        /* branch c1_c1_1_c3_c3 */
                        /* unknown, but legal (optional) field of some other type */

                        res = yami_skip_field(buffer, buffer_size, buffer_position, type);
                    }
                }

                if (res != yami_ok)
                {
                    /* branch c1_c1_1_c4 */
                    break;
                }
            }
        }
        else
        {
            /* branch c1_c2 */
            res = yami_unexpected_value;
        }
    }
    
    return res;
}

/* helper, does not check for buffer size */
static void yami_put_int32_t(uint8_t buffer[], size_t * const buffer_position,
    const int32_t value)
{
    uint8_t byte0;
    uint8_t byte1;
    uint8_t byte2;
    uint8_t byte3;
    
    uint32_t uvalue;
    if (value >= 0)
    {
        /* branch c1 */
        uvalue = (uint32_t)value;
    }
    else
    {
        /* branch c2 */
        const int32_t tmp = -1 * (value + 1);
        uvalue = (uint32_t)tmp ^ 0xffffffffu;
    }
    
    byte0 = (uint8_t)(uvalue & 0xffu);
    byte1 = (uint8_t)((uvalue & 0xff00u) >> 8u);
    byte2 = (uint8_t)((uvalue & 0xff0000u) >> 16u);
    byte3 = (uint8_t)((uvalue & 0xff000000u) >> 24u);
    
    buffer[*buffer_position] = byte0;
    buffer[*buffer_position + 1u] = byte1;
    buffer[*buffer_position + 2u] = byte2;
    buffer[*buffer_position + 3u] = byte3;

    *buffer_position += 4u;
}

/* helper, does not check for buffer size */
static void yami_get_int32_t(const uint8_t buffer[], size_t * const buffer_position,
    int32_t * const value)
{
    const uint8_t byte0 = buffer[*buffer_position];
    const uint8_t byte1 = buffer[*buffer_position + 1u];
    const uint8_t byte2 = buffer[*buffer_position + 2u];
    const uint8_t byte3 = buffer[*buffer_position + 3u];

    const uint32_t uvalue =
        ((uint32_t)byte3 << 24) |
        ((uint32_t)byte2 << 16) |
        ((uint32_t)byte1 << 8) |
        (uint32_t)byte0;
    
    if ((uvalue & 0x80000000u) == 0u)
    {
        /* branch c1 */
        /* positive number */
        
        *value = (int32_t)uvalue;
    }
    else
    {
        /* branch c2 */
        /* negative number */
        
        const uint32_t temp = uvalue ^ 0xffffffffu;
        *value = -1 * (int32_t)temp - 1;
    }
    
    *buffer_position += 4u;
}

void yami_fill_frame_header(uint8_t buffer[],
    const uint32_t message_id,
    const size_t message_header_size, const size_t frame_payload_size)
{
    /* frame header contains 4 int fields: */
    /* - message id                        */
    /* - frame id, always -1               */
    /* - message header size               */
    /* - frame payload size                */

    size_t buffer_position = 0u;

    yami_put_int32_t(buffer, &buffer_position, (int32_t)message_id);
    yami_put_int32_t(buffer, &buffer_position, -1);
    yami_put_int32_t(buffer, &buffer_position, (int32_t)message_header_size);
    yami_put_int32_t(buffer, &buffer_position, (int32_t)frame_payload_size);
}

enum yami_result yami_parse_frame_header(const uint8_t buffer[], const size_t buffer_size,
    size_t * const buffer_position,
    uint32_t * const message_id,
    size_t * const message_header_size, size_t * const frame_payload_size)
{
    const size_t space_left = buffer_size - *buffer_position;
    
    enum yami_result res;
    
    if (space_left < FRAME_HEADER_SIZE)
    {
        /* branch c1 */
        res = yami_not_enough_space;
    }
    else
    {
        /* branch c2 */
        int32_t message_id_tmp;
        int32_t frame_id;
        int32_t message_header_size_tmp;
        int32_t frame_payload_size_tmp;
        
        yami_get_int32_t(buffer, buffer_position, &message_id_tmp);
        yami_get_int32_t(buffer, buffer_position, &frame_id);
        yami_get_int32_t(buffer, buffer_position, &message_header_size_tmp);
        yami_get_int32_t(buffer, buffer_position, &frame_payload_size_tmp);
        
        /* decision frame_header_is_valid: (frame_id == -1) && (message_id_tmp >= 0) && (message_header_size_tmp > 0) && (frame_payload_size_tmp > 0) */
        /*                              A: frame_id == -1              */
        /*                              B: message_id_tmp >= 0         */
        /*                              C: message_header_size_tmp > 0 */
        /*                              D: frame_payload_size_tmp > 0  */
        /* end                                                         */
        if ((frame_id == -1) && (message_id_tmp >= 0) &&
            (message_header_size_tmp > 0) && (frame_payload_size_tmp > 0))
        {
            /* branch c2_c1 */
            *message_id = (uint32_t)message_id_tmp;
            *message_header_size = (uint32_t)message_header_size_tmp;
            *frame_payload_size = (uint32_t)frame_payload_size_tmp;
            
            res = yami_ok;
        }
        else
        {
            /* branch c2_c2 */
            /* this version of YAMI4 does not support multi-frame messages */
            /* and negative payload values are strange anyway              */
            
            res = yami_unexpected_value;
        }
    }
    
    return res;
}
