/* Copyright Inspirel Sp. z o.o. 2013-2015.
 * This file is part of YAMI4.
 *
 * YAMI4 is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * YAMI4 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with YAMI4.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef YAMI4INDUSTRY_SERIALIZATION_H
#define YAMI4INDUSTRY_SERIALIZATION_H

#include "common.h"

#include <stddef.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/** \brief Places type indicator value in the serialization buffer.
 *
 * Places type indicator value in the serialization buffer
 * at the specified index and advances the index to the position
 * immediately after the written value.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param type type code to be written
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_type(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type type);

/** \brief Reads type code from the serialization buffer.
 *
 * Reads type code from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param type (out) type code to be read
 *
 * @return <code>yami_ok</code> or <code>yami_unexpected_value</code>
 *         or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_type(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type * type);

/** \brief Places integer value in the serialization buffer.
 *
 * Places integer value in the serialization buffer at the specified index and
 * advances the index to the position immediately after the written value.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value integer value to be written
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_integer(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value);

/** \brief Reads integer value from the serialization buffer.
 *
 * Reads integer value from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) integer value to be read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_integer(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t * value);

/** \brief Places long long value in the serialization buffer.
 *
 * Places long long value in the serialization buffer at the specified index and
 * advances the index to the position immediately after the written value.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value long long value to be written
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_long_long(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t value);

/** \brief Reads long long value from the serialization buffer.
 *
 * Reads long long value from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) long long value to be read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_long_long(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t * value);

#ifdef YAMI4_WITH_DOUBLE_FLOAT

/** \brief Places double float value in the serialization buffer.
 *
 * Places double float value in the serialization buffer at the specified index and
 * advances the index to the position immediately after the written value.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value double float value to be written
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_double_float(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value);

/** \brief Reads double float value from the serialization buffer.
 *
 * Reads double float value from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) double float value to be read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_double_float(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double * value);

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

/** \brief Places string value in the serialization buffer.
 *
 * Places string value in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: It is assumed that the string is nul-terminated and the function
 *       will stop copying characters at the first nul character.
 *       The function stores the length prefix before the string itself;
 *       the length does not take into account the nul character.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value string value to be written, nul-terminated
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_cstring(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const char value[]);

/** \brief Reads raw string value from the serialization buffer.
 *
 * Reads raw string value from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The string is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output, together with the nul character
 *       that is appended after the extracted value.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the string
 * @param value_length length of string to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_cstring(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    char value[], size_t value_length);

/** \brief Places binary value in the serialization buffer.
 *
 * Places binary value in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: The function stores the length prefix before the binary itself.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value binary value to be written
 * @param value_length length of the binary value
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_binary(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const uint8_t value[], size_t value_length);

/** \brief Reads raw binary value from the serialization buffer.
 *
 * Reads raw binary value from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The binary is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the binary value
 * @param value_length length of binary to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_binary(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    uint8_t value[], size_t value_length);

/** \brief Places boolean array in the serialization buffer.
 *
 * Places boolean array in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: The function stores the length prefix before the array itself.
 *       The boolean array is stored in the packed form, properly padded.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value boolean array to be written,
 *              each array field contains a single boolean value
 * @param value_length length of the array
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_boolean_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int32_t value[], size_t value_length);

/** \brief Reads raw boolean array from the serialization buffer.
 *
 * Reads raw boolean array from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The array is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output.
 *       The array is read in an unpacked form.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the array value
 * @param value_length length of array to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_boolean_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value[], size_t value_length);

/** \brief Places integer array in the serialization buffer.
 *
 * Places integer array in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: The function stores the length prefix before the array itself.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value integer array to be written
 * @param value_length length of the array
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_integer_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int32_t value[], size_t value_length);

/** \brief Reads raw integer array from the serialization buffer.
 *
 * Reads raw integer array from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The array is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the array value
 * @param value_length length of array to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_integer_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int32_t value[], size_t value_length);

/** \brief Places long long array in the serialization buffer.
 *
 * Places long long array in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: The function stores the length prefix before the array itself.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value long long array to be written
 * @param value_length length of the array
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_long_long_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const int64_t value[], size_t value_length);

/** \brief Reads raw long long array from the serialization buffer.
 *
 * Reads raw long long array from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The array is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the array value
 * @param value_length length of array to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_long_long_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t value[], size_t value_length);

#ifdef YAMI4_WITH_DOUBLE_FLOAT

/** \brief Places double float array in the serialization buffer.
 *
 * Places double float array in the serialization buffer at the specified index and
 * advances the index to the position after the written value.
 * Note: The function stores the length prefix before the array itself.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param value double float array to be written
 * @param value_length length of the array
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_put_double_float_array(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const double value[], size_t value_length);

/** \brief Reads raw double float array from the serialization buffer.
 *
 * Reads raw double float array from the serialization buffer from the specified index and
 * advances the index to the position immediately after the value that was read.
 * Note: The array is read without the length prefix, so it is assumed that the caller
 *       has already read the length prefix and has allocated the buffer
 *       that is big enough to accommodate the output.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the value should be read from
 * @param value (out) buffer to store the array value
 * @param value_length length of array to read
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_get_raw_double_float_array(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    double value[], size_t value_length);

#endif /* YAMI4_WITH_DOUBLE_FLOAT */

/** \brief Skips the given field in serialization buffer.
 *
 * Skips the field of the given type in the serialization buffer.
 * This function can be particularly useful when processing fields
 * that can have varying length (strings, binaries and arrays).
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 * @param type type of the field to be skipped
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_skip_field(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    enum yami_parameter_type type);

/** \brief Places message header fields in the serialization buffer.
 *
 * Places message header fields in the serialization buffer at the specified index and
 * advances the index to the position after the header.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param object_name target object name
 * @param message_name target message name
 * @param message_id message identifier
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_fill_message_header(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const char object_name[], const char message_name[], int64_t message_id);

/** \brief Places reply header fields in the serialization buffer.
 *
 * Places reply header fields in the serialization buffer at the specified index and
 * advances the index to the position after the header.
 * This function assumes that the buffer is big enough
 * and does not check the buffer size.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param message_id message identifier
 *
 */
void yami_fill_reply_header(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    int64_t message_id);

/** \brief Places exception header fields in the serialization buffer.
 *
 * Places exception header fields in the serialization buffer at the specified index and
 * advances the index to the position after the header.
 *
 * @param buffer (in/out) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the new value should be written
 * @param reason exception reason
 * @param message_id message identifier
 *
 * @return <code>yami_ok</code> or <code>yami_not_enough_space</code>
 */
enum yami_result yami_fill_exception_header(uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    const char reason[], int64_t message_id);

/** \brief Parses message header.
 *
 * Parses message header and extracts individual fields.
 * Note: only those fields which are found in the header are returned,
 *       the remaining parameters are set to empty strings or to 0.
 *
 * @param buffer (in) serialization buffer
 * @param buffer_size size of the serialization buffer
 * @param buffer_position (in/out) position (in bytes) in the buffer array
 *                        where the header values should be read from
 * @param out_header_type set to the value of "type" field, this field is expected
 * @param header_type_size max length of the header type (char buffer size)
 * @param out_object_name set to the value of "object_name" field, if found
 * @param object_name_size max length of object name (char buffer size)
 * @param out_message_name set to the value of "message_name" field, if found
 * @param message_name_size max length of message name (char buffer size)
 * @param out_reason set to the value of "reason" field, if found
 * @param reason_size max length of reason (char buffer size)
 * @param out_message_id set to the value of "message_id" field, this field is expected
 *
 * @return <code>yami_ok</code> if the header was parsed correctly,
 *         <code>yami_no_such_name</code> if the header lacks expected fields or
 *         <code>yami_not_enough_space</code> if the output parameters are too small or
 *         <code>yami_unexpected_value</code> if the buffer contains unrecognized data
 */
enum yami_result yami_parse_message_header(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    char out_header_type[], size_t header_type_size,
    char out_object_name[], size_t object_name_size,
    char out_message_name[], size_t message_name_size,
    char out_reason[], size_t reason_size,
    int64_t * out_message_id);

void yami_fill_frame_header(uint8_t buffer[],
    uint32_t message_id, size_t message_header_size, size_t frame_payload_size);

enum yami_result yami_parse_frame_header(const uint8_t buffer[], size_t buffer_size,
    size_t * buffer_position,
    uint32_t * message_id,
    size_t * message_header_size, size_t * frame_payload_size);

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* YAMI4INDUSTRY_SERIALIZATION_H */
